"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import logging
import re
import urllib.parse
from datetime import date, datetime
from http.cookies import SimpleCookie
from random import randbytes
from sys import version_info
from time import time
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Unpack, cast

import aiohttp
from bs4 import BeautifulSoup
from yarl import URL as URL_

from . import errors, utils
from .__metadata__ import __version__
from ._const import HTML_PARSER, JSON_DUMPS, JSON_LOADS, URL
from .enums import Currency, Language, Result, Type
from .id import CLAN_ID64_FROM_URL_REGEX, ID, parse_id64
from .models import PriceOverviewDict, api_route
from .types.id import ID32, ID64, AppID, AssetID, BundleID, ChatGroupID, ChatID, PackageID, PostID, TradeOfferID

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Iterable, Mapping, Sequence, ValuesView

    from .client import Client, ClientKwargs
    from .media import Media
    from .types import achievement, app, bundle, clan, guard, trade, user
    from .types.http import AddWalletCode, CMList, Coro, EResultSuccess, ResponseDict, StrOrURL
    from .types.user import IndividualID
    from .user import ClientUser


T = TypeVar("T")
log = logging.getLogger(__name__)


async def json_or_text(r: aiohttp.ClientResponse, *, loads: Callable[[str], Any] = JSON_LOADS) -> Any:
    text = await r.text()
    try:
        if "application/json" in r.headers["Content-Type"]:
            return loads(text)
    except KeyError:
        pass
    return text


class HTTPClient:
    """The HTTP Client that interacts with the Steam web API."""

    def __init__(self, client: Client, **options: Unpack[ClientKwargs]):
        self._session: aiohttp.ClientSession = None  # type: ignore  # filled in login
        self.user: ClientUser = None  # type: ignore
        self._client = client

        self.api_key: str | None = None
        self.language: Language = options.get("language", Language.English)

        self.logged_in = False
        self.user_agent = (
            f"steam.py/{__version__} client (https://github.com/Gobot1234/steam.py), "
            f"Python/{version_info.major}.{version_info.minor}, aiohttp/{aiohttp.__version__}"
        )

        self.proxy: str | None = options.get("proxy")
        self.proxy_auth: aiohttp.BasicAuth | None = options.get("proxy_auth")
        self.connector: aiohttp.BaseConnector | None = options.get("connector")
        self.ssl = options.get("ssl")

    def clear(self) -> None:
        self._session = aiohttp.ClientSession(
            connector=self.connector,
            json_serialize=JSON_DUMPS,
        )

    async def request(
        self,
        method: str,
        url: StrOrURL,
        /,
        api_needs_auth: bool = True,
        supports_access_token: bool = True,
        **kwargs: Any,
    ) -> Any:  # adapted from d.py
        kwargs["headers"] = {"User-Agent": self.user_agent, **kwargs.get("headers", {})}
        payload = kwargs.get("data")

        url = url if isinstance(url, URL_) else URL_(url)

        if url.host == URL.API.host:
            if api_needs_auth:
                kwargs["params"] |= (
                    {"access_token": await self._client._state.ws.access_token()}
                    if self._client._state.login_complete.is_set() and supports_access_token
                    else {"key": await self.get_api_key()}
                )

        elif url.host in (URL.COMMUNITY.host, URL.STORE.host, URL.HELP.host):
            await self.ensure_logged_in()

        r = data = None
        for tries in range(5):
            async with self._session.request(
                method, url, **kwargs, proxy=self.proxy, proxy_auth=self.proxy_auth, ssl=self.ssl
            ) as r:  # noqa: F811
                log.debug("%s %s with PAYLOAD: %s has returned %d", method, r.url, payload, r.status)

                # even errors have text involved in them so this is safe to call
                data = await json_or_text(r)

                # the request was successful so just return the text/json
                if 200 <= r.status < 300:
                    log.debug("%s %s has received %s", method, r.url, data)
                    return data

                # we are being rate limited
                elif r.status == 429:
                    try:
                        # I haven't been able to get any X-Retry-After headers from the API, but we should probably still
                        # handle it
                        delay = float(r.headers["X-Retry-After"])
                    except KeyError:  # steam being un-helpful as usual
                        delay = 2**tries
                    log.warning("We are being rate limited sleeping for %s seconds", delay)
                    await asyncio.sleep(delay)
                    continue

                # we've received a 500 or 502, an unconditional retry
                elif r.status in {500, 502}:
                    await asyncio.sleep(1 + tries * 3)
                    continue

                elif r.status == 401:
                    if "key" in kwargs.get("params", ()) and (
                        isinstance(data, str)
                        and "Access is denied. Retrying will not help. Please verify your <pre>key=</pre>" in data
                    ):  # api key either got revoked or it was never valid, time to fetch a new key
                        key = await self.get_api_key()
                        assert key is not None
                        kwargs["params"]["key"] = key
                        continue  # retry with our new key
                    raise errors.HTTPException(r, data)

                # the usual error cases
                elif r.status == 403:
                    raise errors.Forbidden(r, data)
                elif r.status == 404:
                    raise errors.NotFound(r, data)
                else:
                    raise errors.HTTPException(r, data)

        assert r is not None
        # we've run out of retries, raise
        raise errors.HTTPException(r, data)

    def get(self, url: StrOrURL, **kwargs: Any) -> Coro[Any]:
        return self.request("GET", url, **kwargs)

    def post(self, url: StrOrURL, **kwargs: Any) -> Coro[Any]:
        return self.request("POST", url, **kwargs)

    async def get_cm_list(self, cell_id: int) -> CMList:
        params = {
            "cellid": cell_id,
            "cmtype": "websockets",
        }
        data: ResponseDict[CMList] = await self.get(
            api_route("ISteamDirectory/GetCMListForConnect"), api_needs_auth=False, params=params
        )
        return data["response"]

    def connect_to_cm(self, cm: str) -> Coro[aiohttp.ClientWebSocketResponse]:
        headers = {"User-Agent": self.user_agent}
        return self._session.ws_connect(  # type: ignore  # aiohttp types issue, fixed upstream
            f"wss://{cm}/cmsocket/", headers=headers, proxy=self.proxy, proxy_auth=self.proxy_auth
        )

    @utils.call_once(wait=True)
    async def login(self) -> None:
        jar = self._session.cookie_jar
        steam_login_secure = urllib.parse.quote(f"{self.user.id64}||{await self._client._state.ws.access_token()}")
        for url in (URL.COMMUNITY, URL.STORE, URL.HELP):
            jar.update_cookies(SimpleCookie(f"steamLoginSecure={steam_login_secure}"), url)
            jar.update_cookies(SimpleCookie(f"sessionid={self.session_id}"), url)
        self.logged_in = True

    async def ensure_logged_in(self) -> None:
        if self.logged_in:
            return
        await self._client._state.login_complete.wait()
        await self.login()

    @utils.call_once(wait=True)
    async def get_api_key(self) -> str | None:
        if self.api_key is not None:
            return self.api_key

        resp = await self.get(URL.COMMUNITY / "dev/apikey")
        if (
            "<h2>Access Denied</h2>" in resp
            or "You must have a validated email address to create a Steam Web API key" in resp
        ):
            raise RuntimeError("You must have a premium Steam account or validated email address to use this method")
        if "<h2>Something is wrong</h2>" in resp:
            raise RuntimeError(
                "Your account does not meet the requirements for a developer API key. Your account requires Steam Guard Mobile Authenticator"
            )
        key_re = re.compile(r"<p>Key: ([0-9A-F]+)</p>")
        if match := key_re.findall(resp):
            self.api_key = match[0]
            return self.api_key

        payload = {
            "domain": "steam.py",
            "agreeToTerms": "agreed",
            "sessionid": self.session_id,
            "Submit": "Register",
        }
        resp = await self.post(URL.COMMUNITY / "dev/registerkey", data=payload)
        self.api_key = key_re.findall(resp)[0]
        return self.api_key

    @utils.cached_property
    def session_id(self) -> str:
        return randbytes(16).hex()

    async def close(self) -> None:
        await self._session.close()

    async def get_user(self, user_id64: ID64) -> user.User | None:
        return await anext(self.get_users((user_id64,)), None)

    async def get_users(self, user_id64s: Iterable[ID64]) -> AsyncGenerator[user.User, None]:
        for data in await asyncio.gather(  # gather all the requests concurrently
            *(
                self.get(
                    api_route("ISteamUser/GetPlayerSummaries", version=2),
                    params={"steamids": ",".join(map(str, sublist))},
                    supports_access_token=False,
                )
                for sublist in utils.as_chunks(user_id64s, 100)
            )
        ):
            data: ResponseDict[dict[Literal["players"], list[user.User]]]
            for user in data["response"]["players"]:
                yield user

    async def get_user_escrow(self, user_id64: ID64, token: str | None) -> user.TradeHoldDurations:
        params = {
            "steamid_target": user_id64,
            "trade_offer_access_token": token if token is not None else "",
        }
        data: ResponseDict[user.TradeHoldDurations] = await self.get(
            api_route("IEconService/GetTradeHoldDurations"), params=params
        )
        return data["response"]

    async def get_friends_ids(self, user_id64: ID64) -> list[ID64]:
        params = {
            "steamid": user_id64,
            "relationship": "friend",
        }
        friends: user.FriendsList = await self.get(api_route("ISteamUser/GetFriendList"), params=params)
        return cast("list[ID64]", [int(friend["steamid"]) for friend in friends["friendslist"]["friends"]])

    async def get_user_clans(self, user_id64: ID64) -> list[ID64]:
        params = {"steamid": user_id64}
        data: user.GetUserGroupList = await self.get(api_route("ISteamUser/GetUserGroupList"), params=params)
        return [parse_id64(group["gid"], type=Type.Clan) for group in data["response"]["groups"]]

    async def get_user_bans(self, *user_id64s: ID64) -> list[user.UserBan]:
        params = {"steamids": ",".join(str(id64) for id64 in user_id64s)}
        data: user.GetPlayerBans = await self.get(api_route("ISteamUser/GetPlayerBans"), params=params)
        return [
            {
                "steamid": ID64(int(ban["SteamId"])),
                "community_banned": ban["CommunityBanned"],
                "vac_banned": ban["VACBanned"],
                "number_of_vac_bans": ban["NumberOfVACBans"],
                "days_since_last_ban": ban["DaysSinceLastBan"],
                "number_of_game_bans": ban["NumberOfGameBans"],
                "economy_ban": ban["EconomyBan"],
            }
            for ban in data["players"]
        ]

    async def get_user_level(self, user_id64: ID64) -> int:
        params = {"steamid": user_id64}
        resp: user.GetSteamLevel = await self.get(api_route("IPlayerService/GetSteamLevel"), params=params)
        return resp["response"]["player_level"]

    async def get_user_badges(self, user_id64: ID64) -> user.UserBadges:
        params = {"steamid": user_id64}
        data: ResponseDict[user.UserBadges] = await self.get(api_route("IPlayerService/GetBadges"), params=params)
        return data["response"]

    async def get_user_community_badge_progress(
        self, user_id64: ID64, badge_id: int
    ) -> list[user.CommunityBadgeProgressQuest]:
        params = {
            "steamid": user_id64,
            "badgeid": badge_id,
        }
        data: ResponseDict[dict[Literal["quests"], list[user.CommunityBadgeProgressQuest]]] = await self.get(
            api_route("IPlayerService/GetCommunityBadgeProgress"), params=params
        )
        return data["response"]["quests"]

    async def get_user_recently_played_apps(self, user_id64: ID64) -> list[app.UserRecentlyPlayedApp]:
        params = {"steamid": user_id64}
        data: ResponseDict[dict[Literal["games"], list[app.UserRecentlyPlayedApp]]] = await self.get(
            api_route("IPlayerService/GetRecentlyPlayedGames"), params=params
        )
        return data["response"]["games"]

    async def get_user_wishlist(self, user_id64: ID64) -> AsyncGenerator[tuple[AppID, app.WishlistApp], None]:
        params = {"p": 0}
        while True:
            resp: dict[AppID, app.WishlistApp] = await self.get(
                URL.STORE / f"wishlist/profiles/{user_id64}/wishlistdata", params=params
            )
            if not resp:  # it's an empty list sometimes lol
                return
            for app_id, data in resp.items():
                yield app_id, data
            params["p"] += 1

    async def get_user_inventory(
        self, user_id64: int, app_id: int, context_id: int, language: Language | None
    ) -> trade.Inventory:
        count = 2000
        ret: trade.Inventory = {"assets": [], "descriptions": [], "last_assetid": 0, "more_items": True}  # type: ignore
        while ret["more_items"]:
            params = {
                "count": count,
                "l": (language or self.language).api_name,
                "start_assetid": ret["last_assetid"],
            }
            resp: trade.Inventory = await self.get(
                URL.COMMUNITY / f"inventory/{user_id64}/{app_id}/{context_id}", params=params
            )
            ret["assets"].extend(resp["assets"])
            ret["descriptions"].extend(resp["descriptions"])
            ret["last_assetid"] = resp.get("last_assetid", 0)
            ret["more_items"] = resp.get("more_items", False)
        return ret

    async def get_user_inventory_info(self, user_id64: ID64) -> ValuesView[user.InventoryInfo]:
        resp = await self.get(URL.COMMUNITY / f"profiles/{user_id64}/inventory")
        soup = BeautifulSoup(resp, "html.parser")
        for script in soup.find_all("script", type="text/javascript"):
            if match := re.search(r"var\s+g_rgAppContextData\s*=\s*(?P<json>{.*?});\s*", script.text):
                break
        else:
            raise ValueError("Could not find inventory info")

        return JSON_LOADS(match["json"]).values()

    async def send_user_gift(
        self, user_id: ID32, asset_id: AssetID, name: str, message: str, closing_note: str, signature: str
    ) -> None:
        payload = {
            "GifteeAccountID": user_id,
            "GifteeEmail": "",
            "GifteeName": name,
            "GiftMessage": message,
            "GiftSentiment": closing_note,
            "GiftSignature": signature,
            "GiftGID": asset_id,
            "SessionID": self.session_id,
        }
        headers = {"Referer": f"{URL.STORE}/checkout/sendgift/{asset_id}"}
        data: EResultSuccess = await self.post(URL.STORE / "checkout/sendgiftsubmit", data=payload, headers=headers)
        if data["success"] != Result.OK:
            raise RuntimeError("Failed to send gift")

    async def get_trade_offers(
        self,
        active_only: bool = True,
        sent: bool = True,
        received: bool = True,
        updated_only: bool = True,
        language: Language | None = None,
    ) -> trade.GetTradeOffers:
        params = {
            "active_only": str(active_only).lower(),
            "get_sent_offers": str(sent).lower(),
            "get_received_offers": str(received).lower(),
            "get_descriptions": "true",
            "cursor": 0,
            "language": (language or self.language).api_name,
        }
        if updated_only:
            try:
                params["time_historical_cutoff"] = self.trades_last_fetched
            except AttributeError:
                pass
        resp: ResponseDict[trade.GetTradeOffers] = await self.get(
            api_route("IEconService/GetTradeOffers"), params=params
        )
        first_page = resp["response"]
        next_cursor = first_page.get("next_cursor", 0)
        current_cursor = 0
        while current_cursor < next_cursor:
            params["cursor"] = next_cursor
            resp = await self.get(api_route("IEconService/GetTradeOffers"), params=params)
            page = resp["response"]
            for key, value in page.items():
                value_in_first_page = first_page[key]
                if isinstance(value_in_first_page, list):
                    assert isinstance(value, list)
                    value_in_first_page += value

            current_cursor = next_cursor
            try:
                next_cursor = page["next_cursor"]
            except KeyError:
                break

        if updated_only:
            self.trades_last_fetched = int(time())

        return first_page

    async def get_trade_history(
        self, limit: int, include_failed: bool, previous_time: int = 0, language: Language | None = None
    ) -> trade.GetTradeOfferHistory:
        params = {
            "max_trades": limit,
            "get_descriptions": "true",
            "include_total": "true",
            "include_failed": str(include_failed).lower(),
            "start_after_time": previous_time,
            "language": (language or self.language).api_name,
        }
        data: ResponseDict[trade.GetTradeOfferHistory] = await self.get(
            api_route("IEconService/GetTradeHistory"), params=params
        )
        return data["response"]

    async def get_trade(self, trade_id: TradeOfferID, language: Language | None = None) -> trade.GetTradeOffer:
        params = {
            "tradeofferid": trade_id,
            "get_descriptions": "true",
            "language": (language or self.language).api_name,
        }
        data: ResponseDict[trade.GetTradeOffer] = await self.get(api_route("IEconService/GetTradeOffer"), params=params)
        return data["response"]

    def accept_user_trade(self, user_id64: ID64, trade_id: TradeOfferID) -> Coro[trade.AcceptTrade]:
        payload = {
            "sessionid": self.session_id,
            "tradeofferid": trade_id,
            "serverid": 1,
            "partner": user_id64,
            "captcha": "",
        }
        headers = {"Referer": str(URL.COMMUNITY / f"tradeoffer/{trade_id}")}
        return self.post(URL.COMMUNITY / f"tradeoffer/{trade_id}/accept", data=payload, headers=headers)

    def _cancel_user_trade(self, trade_id: TradeOfferID, option: str) -> Coro[None]:
        payload = {"sessionid": self.session_id}
        return self.post(URL.COMMUNITY / f"tradeoffer/{trade_id}/{option}", data=payload)

    def decline_user_trade(self, trade_id: TradeOfferID) -> Coro[None]:
        return self._cancel_user_trade(trade_id, "decline")

    def cancel_user_trade(self, trade_id: TradeOfferID) -> Coro[None]:
        return self._cancel_user_trade(trade_id, "cancel")

    def send_trade_offer(
        self,
        user: IndividualID,
        sending: list[trade.AssetToDict],
        receiving: list[trade.AssetToDict],
        token: str | None,
        offer_message: str,
        **kwargs: Any,
    ) -> Coro[trade.TradeOfferCreateResponse]:
        payload = {
            "sessionid": self.session_id,
            "serverid": 1,
            "partner": user.id64,
            "tradeoffermessage": offer_message,
            "json_tradeoffer": JSON_DUMPS(
                {
                    "newversion": True,
                    "version": len(sending) + len(receiving) + 1,
                    "me": {"assets": sending, "currency": [], "ready": False},
                    "them": {"assets": receiving, "currency": [], "ready": False},
                }
            ),
            "captcha": "",
            "trade_offer_create_params": JSON_DUMPS({"trade_offer_access_token": token}) if token is not None else "{}",
            **kwargs,
        }
        referer = URL.COMMUNITY / "tradeoffer/new/" % {"partner": str(user.id)}
        if token is not None:
            referer %= {"token": token}
        headers = {"Referer": str(referer)}
        return self.post(URL.COMMUNITY / "tradeoffer/new/send", data=payload, headers=headers)

    async def get_trade_receipt(self, trade_id: int, language: Language | None = None) -> trade.TradeStatus:
        params = {
            "tradeid": trade_id,
            "get_descriptions": "true",
            "language": (language or self.language).api_name,
        }
        data: ResponseDict[trade.TradeStatus] = await self.get(api_route("IEconService/GetTradeStatus"), params=params)
        return data["response"]

    async def check_availability(self, value: str, type: str) -> bool:
        data = {
            "xml": 1,
            "type": type,
            "value": value,
        }
        xml = await self.post(URL.COMMUNITY / "actions/AvailabilityCheck", data=data)
        soup = BeautifulSoup(xml, "xml")
        return soup.response.bResults.text == "1"  # type: ignore

    async def create_clan(
        self,
        name: str,
        abbreviation: str | None,
        community_url_path: str | None,
        public: bool,
    ) -> ID64:
        if not await self.check_availability(name, "groupName"):
            raise ValueError("Name is not available")
        abbreviation = abbreviation or name
        if not 0 < len(abbreviation) < 12:
            raise ValueError("Abbreviation must be between 1 and 12 characters")
        if not await self.check_availability(abbreviation, "abbreviation"):
            raise ValueError("Abbreviation is not available")
        community_url_path = community_url_path or name
        if not await self.check_availability(community_url_path, "groupLink"):
            raise ValueError("Community URL path is not available")

        data = {
            "sessionID": self.session_id,
            "step": 2,  # might need to be 1 then 2
            "groupName": name,
            "abbreviation": abbreviation or name,
            "groupLink": community_url_path or name,
            "bIsPublic": int(public),
        }
        edit_page = await self.post(URL.COMMUNITY / "actions/GroupCreate", data=data)
        soup = BeautifulSoup(edit_page, "html.parser")
        for element in soup.find_all("div", class_="formRow"):
            row_title = element.find("div", class_="formRowTitle")
            if row_title and "ID" in row_title.text.strip():
                return parse_id64(element.find("div", class_="formRowFields").text.strip(), type=Type.Clan)
        raise RuntimeError("Could not find ID should be unreachable")

    async def edit_clan(
        self,
        clan_id64: ID64,
        *,
        abbreviation: str | None = None,
        headline: str | None = None,
        summary: str | None = None,
        community_url_path: str | None = None,
        language: Language | None = None,
        country: str | None = None,
        state: str | None = None,
        city: str | None = None,
        apps: Iterable[AppID] | None = None,
    ) -> None:
        data = {
            "sessionID": self.session_id,
            "type": "profileSave",
            "abbreviation": abbreviation or "",
            "headline": headline or "",
            "summary": summary or "",
            "customURL": community_url_path or "",
            "language": language.api_name if language is not None else "",
            "country": country or "",
            "state": state or "",
            "city": city or "",
            "favorite_games": ",".join(map(str, apps)) if apps is not None else "",
        }
        return await self.post(f"{ID(clan_id64).community_url}/edit", data=data)

    def join_clan(self, clan_id64: ID64) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "action": "join",
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}", data=payload)

    def leave_clan(self, clan_id64: ID64) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "action": "leaveGroup",
            "groupId": clan_id64,
        }
        return self.post(URL.COMMUNITY / "my/home_process", data=payload)

    def invite_user_to_clan(self, user_id64: ID64, clan_id64: ID64) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "group": clan_id64,
            "invitee": user_id64,
            "type": "groupInvite",
        }
        return self.post(URL.COMMUNITY / "actions/GroupInvite", data=payload)

    def clear_nickname_history(self) -> Coro[None]:
        payload = {"sessionid": self.session_id}
        return self.post(URL.COMMUNITY / "my/ajaxclearaliashistory", data=payload)

    def get_price(self, app_id: AppID, item_name: str, currency: Currency | None) -> Coro[PriceOverviewDict]:
        params = {
            "appid": app_id,
            "market_hash_name": item_name,
        }
        if currency is not None:
            params |= {"currency": currency}

        return self.get(URL.COMMUNITY / "market/priceoverview", params=params)

    async def get_clan_members(self, clan_id64: ID64) -> AsyncGenerator[ID32, None]:
        url = f"{ID(clan_id64).community_url}/members"
        page = 1
        number_of_pages = None

        while number_of_pages is None or page <= number_of_pages:
            resp = await self.get(url, params={"p": page, "content_only": "true"})
            soup = BeautifulSoup(resp, HTML_PARSER)
            if not number_of_pages:
                page_select = soup.find("div", class_="group_paging")
                assert page_select is not None
                number_of_pages = int(re.findall(r"\d* - (\d*)", page_select.text)[0])

            for s in soup.find_all("div", id="memberList"):
                for user in s.find_all("div", class_="member_block"):
                    yield int(user["data-miniprofile"])  # type: ignore
            page += 1

    async def get_clan_invitees(self) -> dict[ID64, ID64]:
        # N.B. can only be invited by one person at a time
        resp = await self.get(URL.COMMUNITY / "my/groups/pending", params={"ajax": "1"})
        soup = BeautifulSoup(resp, HTML_PARSER)
        elements = soup.find_all("a", class_="linkStandard")

        return {
            ID64(
                int(
                    CLAN_ID64_FROM_URL_REGEX.search(clan_element["href"])["steamid"],
                )
            ): parse_id64(invitee_element["data-miniprofile"])
            for (clan_element, invitee_element) in zip(
                (element for element in elements if "steamLink" in element["class"]),
                (element for element in elements if "data-miniprofile" in element.attrs),
            )
        }

    async def get_clan_announcement_ids(self, clan_id64: ID64) -> list[int]:
        rss = await self.get(URL.COMMUNITY / f"gid/{clan_id64}/rss")
        soup = BeautifulSoup(rss, HTML_PARSER)
        return [
            int(match[0])
            for url in soup.find_all("guid")
            if (match := re.findall(r"announcements/detail/(\d+)", url.text))
        ]

    async def get_clan_events_for(self, clan_id64: ID64, date: date) -> list[int]:
        xml = await self.post(
            URL.COMMUNITY / f"gid/{clan_id64}/events",
            data={"xml": 1, "action": "eventFeed", "month": date.month, "year": date.year},
        )
        soup = BeautifulSoup(xml, HTML_PARSER)
        return [
            int(url.rpartition("/")[2])
            for event_title in soup.find_all("div", class_="eventBlockTitle")
            for url in event_title.a.get("href")
            if url
        ]

    def _edit_clan_event(
        self,
        action: str,
        clan_id64: ID64,
        name: str,
        description: str,
        event_type: str,
        app_id: str,
        server_ip: str,
        server_password: str,
        start: datetime | None,
        event_id: int | None,
    ) -> Coro[str]:
        if start is None:
            tz_offset = int(datetime.now().astimezone().tzinfo.utcoffset(None).total_seconds())  # type: ignore  # PEP 696 should solve this in typeshed
            start_date = "MM/DD/YY"
            start_hour = "12"
            start_minute = "00"
            start_ampm = "PM"
            time_choice = "quick"
        else:
            if start.tzinfo is None:
                start = start.astimezone()
            assert start.tzinfo is not None
            tz_offset = int(start.tzinfo.utcoffset(None).total_seconds())  # type: ignore

            start_date = f"{start:%m/%d/%y}"
            start_hour = f"{start:%I}"
            start_minute = f"{start:%M}"
            start_ampm = "AM" if start.hour <= 12 else "PM"
            time_choice = "specific"

        data = {
            "sessionid": self.session_id,
            "action": action,
            "tzOffset": tz_offset,
            "name": name,
            "type": event_type,
            "appID": app_id,
            "serverIP": server_ip,
            "serverPassword": server_password,
            "notes": description,
            "eventQuickTime": "now",
            "startDate": start_date,
            "startHour": start_hour,
            "startMinute": start_minute,
            "startAMPM": start_ampm,
            "timeChoice": time_choice,
        }

        if event_id is not None:
            data["eventID"] = event_id

        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/eventEdit", data=data)

    def create_clan_event(self, *args: Any, **kwargs: Any) -> Coro[str]:
        return self._edit_clan_event("newEvent", *args, event_id=None, **kwargs)

    def edit_clan_event(self, *args: Any, **kwargs: Any) -> Coro[str]:
        return self._edit_clan_event("updateEvent", *args, **kwargs)

    def delete_clan_event(self, clan_id64: ID64, event_id: int) -> Coro[None]:
        data = {
            "sessionid": self.session_id,
            "action": "deleteEvent",
            "eventID": event_id,
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/events", data=data)

    async def get_clan_events(self, clan_id: int, event_ids: Sequence[int]) -> list[clan.Event]:
        params = {
            "clanid_list": ",".join([str(clan_id)] * len(event_ids)),
            "uniqueid_list": ",".join(str(id) for id in event_ids),
        }
        data: dict[Literal["events"], list[clan.Event]] = await self.get(
            URL.STORE / "events/ajaxgeteventdetails", params=params
        )
        return data["events"]

    def create_clan_announcement(
        self, clan_id64: ID64, name: str, description: str, hidden: bool = False
    ) -> Coro[None]:
        data = {
            "sessionID": self.session_id,
            "action": "post",
            "headline": name,
            "body": description,
            "languages[0][headline]": name,
            "languages[0][body]": description,
        }
        if hidden:
            data["is_hidden"] = "is_hidden"
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/announcements", data=data)

    def edit_clan_announcement(self, clan_id64: ID64, announcement_id: int, name: str, description: str) -> Coro[None]:
        data = {
            "sessionID": self.session_id,
            "gid": announcement_id,
            "action": "update",
            "headline": name,
            "body": description,
            "languages[0][headline]": name,
            "languages[0][body]": description,
            "languages[0][updated]": 1,
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/announcements", data=data)

    def delete_clan_announcement(self, clan_id64: ID64, announcement_id: int) -> Coro[None]:
        params = {
            "sessionID": self.session_id,
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/announcements/delete/{announcement_id}", params=params)

    async def get_clan_announcement(
        self,
        clan_id: int,
        announcement_id: int,
    ) -> clan.Event:
        params = {
            "clan_accountid": clan_id,
            "announcement_gid": announcement_id,
        }
        data: clan.GetClanAnnouncement = await self.get(URL.STORE / "events/ajaxgetpartnerevent", params=params)
        return data["event"]

    def vote_on_user_post(self, user_id64: ID64, post_id: PostID, vote: int) -> Coro[None]:
        data = {
            "sessionid": self.session_id,
            "vote": vote,
        }

        return self.post(URL.COMMUNITY / f"comment/UserStatusPublished/voteup/{user_id64}/{post_id}", data=data)

    def post_review(
        self,
        app_id: AppID,
        content: str,
        upvoted: bool,
        public: bool,
        commentable: bool,
        received_compensation: bool,
        language: str,
    ) -> Coro[None]:
        data = {
            "appid": app_id,
            "steamworksappid": app_id,
            "comment": content,
            "rated_up": str(upvoted).lower(),
            "is_public": str(public).lower(),
            "language": language,
            "received_compensation": int(received_compensation),
            "disable_comments": int(not commentable),
            "sessionid": self.session_id,
        }

        return self.post(URL.STORE / "friends/recommendgame", data=data)

    def get_reviews(
        self, app_id: AppID, filter: str, review_type: str, purchase_type: str, cursor: str = "*"
    ) -> Coro[dict[str, Any]]:
        params = {
            "json": 1,
            "num_per_page": 100,
            "cursor": urllib.parse.quote(cursor),
            "filter": filter,
            "review_type": review_type,
            "purchase_type": purchase_type,
        }

        return self.get(URL.STORE / f"appreviews/{app_id}", params=params)

    def mark_review_as_helpful(self, review_id: int, rated_up: bool) -> Coro[None]:
        data = {
            "rateup": str(rated_up).lower(),
            "sessionid": self.session_id,
        }
        return self.post(URL.COMMUNITY / f"userreviews/rate/{review_id}", data=data)

    def mark_review_as_funny(self, review_id: int) -> Coro[None]:
        data = {
            "tagid": 1,
            "rateup": "true",
            "sessionid": self.session_id,
        }
        return self.post(URL.COMMUNITY / f"userreviews/votetag/{review_id}", data=data)

    def delete_review(self, app_id: AppID) -> Coro[None]:
        data = {
            "action": "delete",
            "appid": app_id,
            "sessionid": self.session_id,
        }
        return self.post(URL.COMMUNITY / "my/recommended", data=data)

    def get_app(self, app_id: AppID, language: Language | None) -> Coro[dict[str, Any]]:
        params = {
            "appids": app_id,
            "l": (language or self.language).api_name,
        }
        return self.get(URL.STORE / "api/appdetails", params=params)

    def get_app_dlc(self, app_id: AppID, language: Language | None) -> Coro[dict[str, Any]]:
        params = {
            "appid": app_id,
            "l": (language or self.language).api_name,
        }
        return self.get(URL.STORE / "api/dlcforapp", params=params)

    async def get_app_asset_prices(self, app_id: AppID, currency: Currency | None = None) -> app.AssetPrices:
        params = {
            "appid": app_id,
        }
        if currency is not None:
            params |= {
                "currency": currency.name,
            }

        data: dict[Literal["result"], app.AssetPrices] = await self.get(
            api_route("ISteamEconomy/GetAssetPrices"), params=params
        )
        return data["result"]

    def get_app_suggestions(
        self, term: str, language: Language | None = None
    ) -> Coro[list[dict[{"name": str, "id": str, "type": str}]]]:  # noqa: UP037, F821
        params = {
            "term": term,
            "f": "json",
            "cc": "en",
            "l": (language or self.language).api_name,
        }
        return self.get(URL.STORE / "search/suggest", params=params)

    async def get_all_apps(
        self,
        include_games: bool,
        include_dlc: bool,
        include_software: bool,
        include_videos: bool,
        include_hardware: bool,
        chunk_size: int | None,
        limit: int | None,
        last_app_id: AppID | None = None,
        modified_after: datetime | None = None,
    ) -> AsyncGenerator[app.AppListApp, None]:
        have_more_results = True
        last_app_id = None
        while have_more_results:
            params = {
                "include_games": str(include_games).lower(),
                "include_dlc": str(include_dlc).lower(),
                "include_software": str(include_software).lower(),
                "include_videos": str(include_videos).lower(),
                "include_hardware": str(include_hardware).lower(),
                "max_results": min(chunk_size if chunk_size is not None else 10_000, 50_000),
            }
            if last_app_id is not None:
                params["last_appid"] = last_app_id
            if modified_after is not None:
                params["if_modified_since"] = int(modified_after.timestamp())
            data = await self.get(api_route("IStoreService/GetAppList"), params=params)
            resp = data["response"]
            for app in resp["apps"]:
                yield app
                if limit is not None:
                    limit -= 1
                    if limit == 0:
                        return
            last_app_id = AppID(resp.get("last_appid", 0))
            have_more_results = resp.get("have_more_results", False)

    async def get_app_stats(self, app_id: AppID, language: Language | None) -> achievement.AppAppStats:
        params = {
            "appid": app_id,
            "l": (language or self.language).api_name,
        }
        data: dict[Literal["game"], achievement.AppAppStats] = await self.get(
            api_route("ISteamUserStats/GetSchemaForGame", 2), params=params
        )
        return data["game"]

    async def get_app_leaderboards(self, app_id: AppID, language: Language | None) -> list[app.Leaderboard]:
        params = {
            "xml": 1,
            "l": (language or self.language).api_name,
        }
        xml = await self.get(URL.COMMUNITY / f"stats/{app_id}/leaderboards", params=params)
        soup = BeautifulSoup(xml, HTML_PARSER)
        assert soup.response is not None

        return [
            {
                "id": int(leaderboard.lbid.text),
                "name": leaderboard.find("name").text,
                "display_name": str(leaderboard.display_name.text),
                "entry_count": int(leaderboard.entries.text),
                "sort_method": int(leaderboard.sortmethod.text),
                "display_type": int(leaderboard.displaytype.text),
            }
            for leaderboard in soup.response.find_all("leaderboard")
        ]

    async def verify_app_ticket(
        self, app_id: AppID, ticket: str, publisher_key: str | None
    ) -> user.AuthenticateUserTicketParams:
        params = {
            "appid": app_id,
            "ticket": ticket,
        }
        if publisher_key is not None:
            params["key"] = publisher_key
        resp: ResponseDict[user.AuthenticateUserTicket] = await self.get(
            api_route("ISteamUserAuth/AuthenticateUserTicket", publisher=publisher_key is not None),
            params=params,
            api_needs_auth=publisher_key is None,
        )
        return resp["response"]["params"]

    def get_package(self, package_id: PackageID, language: Language | None) -> Coro[dict[str, Any]]:
        params = {
            "packageids": package_id,
            "l": (language or self.language).api_name,
        }
        return self.get(URL.STORE / "api/packagedetails", params=params)

    def redeem_package(self, package_id: PackageID) -> Coro[dict[str, Any]]:
        data = {
            "ajax": "true",
            "sessionid": self.session_id,
        }
        return self.post(URL.STORE / f"freelicense/addfreelicense/{package_id}", data=data)

    def remove_license(self, license_id: PackageID) -> Coro[None]:
        data = {
            "sessionid": self.session_id,
            "packageid": license_id,
        }
        return self.post(URL.STORE / "account/removelicense", data=data)

    def get_bundle(self, bundle_id: BundleID, language: Language | None) -> Coro[list[bundle.Bundle]]:
        params = {
            "bundleids": bundle_id,
            "l": (language or self.language).api_name,
            "cc": "en",
        }
        return self.get(URL.STORE / "actions/ajaxresolvebundles", params=params)

    def redeem_bundle(self, bundle_id: BundleID) -> Coro[dict[str, Any]]:
        data = {
            "ajax": "true",
            "sessionid": self.session_id,
        }
        return self.post(URL.STORE / f"freelicense/addfreebundle/{bundle_id}", data=data)

    def add_wallet_code(self, code: str) -> Coro[AddWalletCode]:
        data = {"wallet_code": code, "sessionid": self.session_id}
        return self.post(URL.STORE / "account/ajaxredeemwalletcode", data=data)

    def add_phone_number(self, phone_number: str) -> Coro[guard.AddPhoneNumber]:
        data = {
            "op": "add_phone_number",
            "arg": phone_number,
            "checkfortos": 0,
            "skipvoip": 0,
            "sessionid": self.session_id,
        }
        return self.post(URL.COMMUNITY / "steamguard/phoneajax", data=data)

    async def edit_profile_info(
        self,
        name: str | None,
        real_name: str | None,
        url: str | None,
        summary: str | None,
        country: str | None,
        state: str | None,
        city: str | None,
    ) -> None:
        info = await self.user.profile_info()

        await self.ensure_logged_in()

        async with self._session.get(URL.COMMUNITY / "my") as r:
            current_url = r.url

        payload = {
            "sessionID": self.session_id,
            "type": "profileSave",
            "weblink_1_title": "",
            "weblink_1_url": "",
            "weblink_2_title": "",
            "weblink_2_url": "",
            "weblink_3_title": "",
            "weblink_3_url": "",
            "personaName": name or self.user.name,
            "real_name": real_name or info.real_name or "",
            "customURL": (url if url is not None else current_url.parts[-2]) or "",
            "country": country or info.country_name,
            "state": state or info.state_name,
            "city": city or info.city_name,
            "summary": summary or info.summary,
            "json": "1",
        }

        await self.post(f"{self.user.community_url}/edit", data=payload)

    async def update_avatar(self, avatar: Media, type: str, *, params: Mapping[str, Any] = {}, **extras: Any) -> None:
        with avatar:
            payload = aiohttp.FormData(fields={k: str(v) for k, v in extras.items()})
            payload.add_field("MAX_FILE_SIZE", str(avatar.size))
            payload.add_field("type", type)
            payload.add_field("sessionid", self.session_id)
            payload.add_field("doSub", "1")
            payload.add_field(
                "avatar", avatar.read(), filename=f"avatar.{avatar.type}", content_type=f"image/{avatar.type}"
            )
            await self.post(URL.COMMUNITY / "actions/FileUploader", data=payload, params=params)

    async def send_media(self, media: Media, **kwargs: int) -> None:
        contents = media.read()
        payload = {
            "sessionid": self.session_id,
            "l": "english",
            "file_size": media.size,
            "file_name": media.name,
            "file_sha": media.hash(contents),
            "file_image_width": media.width,
            "file_image_height": media.height,
            "file_type": media.type,
        }
        resp = await self.post(URL.COMMUNITY / "chat/beginfileupload", data=payload)

        result = resp["result"]
        url = f'{"https" if result["use_https"] else "http"}://{result["url_host"]}{result["url_path"]}'
        headers = {header["name"]: header["value"] for header in result["request_headers"]}
        await self.request("PUT", url, headers=headers, data=contents)

        payload |= {
            "success": 1,
            "ugcid": result["ugcid"],
            "timestamp": result["timestamp"],
            "hmac": resp["hmac"],
            "spoiler": int(media.spoiler),
        } | kwargs
        await self.post(URL.COMMUNITY / "chat/commitfileupload", data=payload)

    async def send_user_media(self, user_id64: ID64, media: Media) -> None:
        with media:
            await self.send_media(media, friend_steamid=user_id64)

    async def send_chat_media(self, chat_group_id: ChatGroupID, chat_id: ChatID, media: Media) -> None:
        with media:
            await self.send_media(media, chat_group_id=chat_group_id, chat_id=chat_id)

    async def upload_chat_icon(self, media: Media) -> bytes:
        with media:
            payload = aiohttp.FormData()
            payload.add_field("sessionid", self.session_id)
            payload.add_field("avatar", media.read(), filename=media.name, content_type=f"image/{media.type}")
            resp = await self.post(URL.COMMUNITY / "chat/avatarfileupload", data=payload)
            return bytes.fromhex(resp["sha"])



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_parental.steamclient.proto
# plugin: python-betterproto
# Last updated 10/08/2022

from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


class EProfileCustomizationType(betterproto.Enum):
    Invalid = 0
    RareAchievementShowcase = 1
    GameCollector = 2
    ItemShowcase = 3
    TradeShowcase = 4
    Badges = 5
    FavoriteGame = 6
    ScreenshotShowcase = 7
    CustomText = 8
    FavoriteGroup = 9
    Recommendation = 10
    WorkshopItem = 11
    MyWorkshop = 12
    ArtworkShowcase = 13
    VideoShowcase = 14
    Guides = 15
    MyGuides = 16
    Achievements = 17
    Greenlight = 18
    MyGreenlight = 19
    Salien = 20
    LoyaltyRewardReactions = 21
    SingleArtworkShowcase = 22
    AchievementsCompletionist = 23


class EBanContentCheckResult(betterproto.Enum):
    NotScanned = 0
    Reset = 1
    NeedsChecking = 2
    VeryUnlikely = 5
    Unlikely = 30
    Possible = 50
    Likely = 75
    VeryLikely = 100


class EProfileCustomizationStyle(betterproto.Enum):
    Default = 0
    Selected = 1
    Rarest = 2
    MostRecent = 3
    Random = 4
    HighestRated = 5


class EAgreementType(betterproto.Enum):
    Invalid = -1
    GlobalSSA = 0
    ChinaSSA = 1


class ENotificationSetting(betterproto.Enum):
    UseDefault = 0
    Always = 1
    Never = 2


class ETextFilterSetting(betterproto.Enum):
    SteamLabOptedOut = 0
    Enabled = 1
    EnabledAllowProfanity = 2
    Disabled = 3


class GetMutualFriendsForIncomingInvitesRequest(UnifiedMessage, um_name="Player.GetMutualFriendsForIncomingInvites"):
    pass


@dataclass(eq=False, repr=False)
class IncomingInviteMutualFriendList(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    mutual_friend_account_ids: list[int] = betterproto.uint32_field(2)


class GetMutualFriendsForIncomingInvitesResponse(UnifiedMessage, um_name="Player.GetMutualFriendsForIncomingInvites"):
    incoming_invite_mutual_friends_lists: "list[IncomingInviteMutualFriendList]" = betterproto.message_field(1)


class GetOwnedGamesRequest(UnifiedMessage, um_name="Player.GetOwnedGames"):
    steamid: int = betterproto.uint64_field(1)
    include_appinfo: bool = betterproto.bool_field(2)
    include_played_free_games: bool = betterproto.bool_field(3)
    appids_filter: list[int] = betterproto.uint32_field(4)
    include_free_sub: bool = betterproto.bool_field(5)
    skip_unvetted_apps: bool = betterproto.bool_field(6)


class GetOwnedGamesResponse(UnifiedMessage, um_name="Player.GetOwnedGames"):
    game_count: int = betterproto.uint32_field(1)
    games: "list[GetOwnedGamesResponseGame]" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class GetOwnedGamesResponseGame(betterproto.Message):
    appid: int = betterproto.int32_field(1)
    name: str = betterproto.string_field(2)
    playtime_2_weeks: int = betterproto.int32_field(3)
    playtime_forever: int = betterproto.int32_field(4)
    img_icon_url: str = betterproto.string_field(5)
    has_community_visible_stats: bool = betterproto.bool_field(7)
    playtime_windows_forever: int = betterproto.int32_field(8)
    playtime_mac_forever: int = betterproto.int32_field(9)
    playtime_linux_forever: int = betterproto.int32_field(10)
    rtime_last_played: int = betterproto.uint32_field(11)
    capsule_filename: str = betterproto.string_field(12)
    sort_as: str = betterproto.string_field(13)
    has_workshop: bool = betterproto.bool_field(14)
    has_market: bool = betterproto.bool_field(15)
    has_dlc: bool = betterproto.bool_field(16)
    has_leaderboards: bool = betterproto.bool_field(17)
    content_descriptorids: list[int] = betterproto.uint32_field(18)


class GetPlayNextRequest(UnifiedMessage, um_name="Player.GetPlayNext"):
    max_age_seconds: int = betterproto.uint32_field(1)
    ignore_appids: list[int] = betterproto.uint32_field(2)


class GetPlayNextResponse(UnifiedMessage, um_name="Player.GetPlayNext"):
    last_update_time: int = betterproto.uint32_field(1)
    appids: list[int] = betterproto.uint32_field(2)


class GetFriendsGameplayInfoRequest(UnifiedMessage, um_name="Player.GetFriendsGameplayInfo"):
    appid: int = betterproto.uint32_field(1)


class GetFriendsGameplayInfoResponse(UnifiedMessage, um_name="Player.GetFriendsGameplayInfo"):
    your_info: "GetFriendsGameplayInfoResponseOwnGameplayInfo" = betterproto.message_field(1)
    in_game: "list[GetFriendsGameplayInfoResponseFriendsGameplayInfo]" = betterproto.message_field(2)
    played_recently: "list[GetFriendsGameplayInfoResponseFriendsGameplayInfo]" = betterproto.message_field(3)
    played_ever: "list[GetFriendsGameplayInfoResponseFriendsGameplayInfo]" = betterproto.message_field(4)
    owns: "list[GetFriendsGameplayInfoResponseFriendsGameplayInfo]" = betterproto.message_field(5)
    in_wishlist: "list[GetFriendsGameplayInfoResponseFriendsGameplayInfo]" = betterproto.message_field(6)


@dataclass(eq=False, repr=False)
class GetFriendsGameplayInfoResponseFriendsGameplayInfo(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    minutes_played: int = betterproto.uint32_field(2)
    minutes_played_forever: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class GetFriendsGameplayInfoResponseOwnGameplayInfo(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    minutes_played: int = betterproto.uint32_field(2)
    minutes_played_forever: int = betterproto.uint32_field(3)
    in_wishlist: bool = betterproto.bool_field(4)
    owned: bool = betterproto.bool_field(5)


class GetFriendsAppsActivityRequest(UnifiedMessage, um_name="Player.GetFriendsAppsActivity"):
    news_language: str = betterproto.string_field(1)
    request_flags: int = betterproto.uint32_field(2)


class GetFriendsAppsActivityResponse(UnifiedMessage, um_name="Player.GetFriendsAppsActivity"):
    trending: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(1)
    recent_purchases: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(2)
    unowned: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(3)
    popular: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(4)
    dont_forget: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(5)
    being_discussed: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(6)
    new_to_group: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(7)
    returned_to_group: "list[GetFriendsAppsActivityResponseAppFriendsInfo]" = betterproto.message_field(8)
    active_friend_count: int = betterproto.uint32_field(9)


@dataclass(eq=False, repr=False)
class GetFriendsAppsActivityResponseFriendPlayTime(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    minutes_played_this_week: int = betterproto.uint32_field(2)
    minutes_played_two_weeks: int = betterproto.uint32_field(3)
    minutes_played_forever: int = betterproto.uint32_field(4)
    event_count: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class GetFriendsAppsActivityResponseAppFriendsInfo(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    friends: "list[GetFriendsAppsActivityResponseFriendPlayTime]" = betterproto.message_field(2)
    display_order: int = betterproto.uint32_field(3)


class GetGameBadgeLevelsRequest(UnifiedMessage, um_name="Player.GetGameBadgeLevels"):
    appid: int = betterproto.uint32_field(1)


class GetGameBadgeLevelsResponse(UnifiedMessage, um_name="Player.GetGameBadgeLevels"):
    player_level: int = betterproto.uint32_field(1)
    badges: "list[GetGameBadgeLevelsResponseBadge]" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class GetGameBadgeLevelsResponseBadge(betterproto.Message):
    level: int = betterproto.int32_field(1)
    series: int = betterproto.int32_field(2)
    border_color: int = betterproto.uint32_field(3)


class GetProfileBackgroundRequest(UnifiedMessage, um_name="Player.GetProfileBackground"):
    steamid: int = betterproto.fixed64_field(1)
    language: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ProfileItem(betterproto.Message):
    communityitemid: int = betterproto.uint64_field(1)
    image_small: str = betterproto.string_field(2)
    image_large: str = betterproto.string_field(3)
    name: str = betterproto.string_field(4)
    item_title: str = betterproto.string_field(5)
    item_description: str = betterproto.string_field(6)
    appid: int = betterproto.uint32_field(7)
    item_type: int = betterproto.uint32_field(8)
    item_class: int = betterproto.uint32_field(9)
    movie_webm: str = betterproto.string_field(10)
    movie_mp4: str = betterproto.string_field(11)
    movie_webm_small: str = betterproto.string_field(13)
    movie_mp4_small: str = betterproto.string_field(14)
    equipped_flags: int = betterproto.uint32_field(12)


class GetProfileBackgroundResponse(UnifiedMessage, um_name="Player.GetProfileBackground"):
    profile_background: "ProfileItem" = betterproto.message_field(1)


class SetProfileBackgroundRequest(UnifiedMessage, um_name="Player.SetProfileBackground"):
    communityitemid: int = betterproto.uint64_field(1)


class SetProfileBackgroundResponse(UnifiedMessage, um_name="Player.SetProfileBackground"):
    pass


class GetMiniProfileBackgroundRequest(UnifiedMessage, um_name="Player.GetMiniProfileBackground"):
    steamid: int = betterproto.fixed64_field(1)
    language: str = betterproto.string_field(2)


class GetMiniProfileBackgroundResponse(UnifiedMessage, um_name="Player.GetMiniProfileBackground"):
    profile_background: "ProfileItem" = betterproto.message_field(1)


class SetMiniProfileBackgroundRequest(UnifiedMessage, um_name="Player.SetMiniProfileBackground"):
    communityitemid: int = betterproto.uint64_field(1)


class SetMiniProfileBackgroundResponse(UnifiedMessage, um_name="Player.SetMiniProfileBackground"):
    pass


class GetAvatarFrameRequest(UnifiedMessage, um_name="Player.GetAvatarFrame"):
    steamid: int = betterproto.fixed64_field(1)
    language: str = betterproto.string_field(2)


class GetAvatarFrameResponse(UnifiedMessage, um_name="Player.GetAvatarFrame"):
    avatar_frame: "ProfileItem" = betterproto.message_field(1)


class SetAvatarFrameRequest(UnifiedMessage, um_name="Player.SetAvatarFrame"):
    communityitemid: int = betterproto.uint64_field(1)


class SetAvatarFrameResponse(UnifiedMessage, um_name="Player.SetAvatarFrame"):
    pass


class GetAnimatedAvatarRequest(UnifiedMessage, um_name="Player.GetAnimatedAvatar"):
    steamid: int = betterproto.fixed64_field(1)
    language: str = betterproto.string_field(2)


class GetAnimatedAvatarResponse(UnifiedMessage, um_name="Player.GetAnimatedAvatar"):
    avatar: "ProfileItem" = betterproto.message_field(1)


class SetAnimatedAvatarRequest(UnifiedMessage, um_name="Player.SetAnimatedAvatar"):
    communityitemid: int = betterproto.uint64_field(1)


class SetAnimatedAvatarResponse(UnifiedMessage, um_name="Player.SetAnimatedAvatar"):
    pass


class GetProfileItemsOwnedRequest(UnifiedMessage, um_name="Player.GetProfileItemsOwned"):
    language: str = betterproto.string_field(1)


class GetProfileItemsOwnedResponse(UnifiedMessage, um_name="Player.GetProfileItemsOwned"):
    profile_backgrounds: "list[ProfileItem]" = betterproto.message_field(1)
    mini_profile_backgrounds: "list[ProfileItem]" = betterproto.message_field(2)
    avatar_frames: "list[ProfileItem]" = betterproto.message_field(3)
    animated_avatars: "list[ProfileItem]" = betterproto.message_field(4)
    profile_modifiers: "list[ProfileItem]" = betterproto.message_field(5)


class GetProfileItemsEquippedRequest(UnifiedMessage, um_name="Player.GetProfileItemsEquipped"):
    steamid: int = betterproto.fixed64_field(1)
    language: str = betterproto.string_field(2)


class GetProfileItemsEquippedResponse(UnifiedMessage, um_name="Player.GetProfileItemsEquipped"):
    profile_background: "ProfileItem" = betterproto.message_field(1)
    mini_profile_background: "ProfileItem" = betterproto.message_field(2)
    avatar_frame: "ProfileItem" = betterproto.message_field(3)
    animated_avatar: "ProfileItem" = betterproto.message_field(4)
    profile_modifier: "ProfileItem" = betterproto.message_field(5)


class SetEquippedProfileItemFlagsRequest(UnifiedMessage, um_name="Player.SetEquippedProfileItemFlags"):
    communityitemid: int = betterproto.uint64_field(1)
    flags: int = betterproto.uint32_field(2)


class SetEquippedProfileItemFlagsResponse(UnifiedMessage, um_name="Player.SetEquippedProfileItemFlags"):
    pass


class GetEmoticonListRequest(UnifiedMessage, um_name="Player.GetEmoticonList"):
    pass


class GetEmoticonListResponse(UnifiedMessage, um_name="Player.GetEmoticonList"):
    emoticons: "list[GetEmoticonListResponseEmoticon]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetEmoticonListResponseEmoticon(betterproto.Message):
    name: str = betterproto.string_field(1)
    count: int = betterproto.int32_field(2)
    time_last_used: int = betterproto.uint32_field(3)
    use_count: int = betterproto.uint32_field(4)
    time_received: int = betterproto.uint32_field(5)
    appid: int = betterproto.uint32_field(6)


class GetTopAchievementsForGamesRequest(UnifiedMessage, um_name="Player.GetTopAchievementsForGames"):
    steamid: int = betterproto.uint64_field(1)
    language: str = betterproto.string_field(2)
    max_achievements: int = betterproto.uint32_field(3)
    appids: list[int] = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class GetTopAchievementsForGamesResponseGameAchievement(betterproto.Message):
    statid: int = betterproto.uint32_field(1)
    bit: int = betterproto.uint32_field(2)
    name: str = betterproto.string_field(3)
    desc: str = betterproto.string_field(4)
    icon: str = betterproto.string_field(5)
    icon_gray: str = betterproto.string_field(6)
    hidden: bool = betterproto.bool_field(7)
    player_percent_unlocked: str = betterproto.string_field(8)


@dataclass(eq=False, repr=False)
class GetTopAchievementsForGamesResponseGame(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    total_achievements: int = betterproto.uint32_field(2)
    achievements: list[GetTopAchievementsForGamesResponseGameAchievement] = betterproto.message_field(3)


class GetTopAchievementsForGamesResponse(UnifiedMessage, um_name="Player.GetTopAchievementsForGames"):
    games: "list[GetTopAchievementsForGamesResponseGame]" = betterproto.message_field(1)


class GetAchievementsProgressRequest(UnifiedMessage, um_name="Player.GetAchievementsProgress"):
    steamid: int = betterproto.uint64_field(1)
    language: str = betterproto.string_field(2)
    appids: list[int] = betterproto.uint32_field(3)


class GetAchievementsProgressResponse(UnifiedMessage, um_name="Player.GetAchievementsProgress"):
    achievement_progress: "list[GetAchievementsProgressResponseAchievementProgress]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetAchievementsProgressResponseAchievementProgress(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    unlocked: int = betterproto.uint32_field(2)
    total: int = betterproto.uint32_field(3)
    percentage: float = betterproto.float_field(4)
    all_unlocked: bool = betterproto.bool_field(5)
    cache_time: int = betterproto.uint32_field(6)


class GetGameAchievementsRequest(UnifiedMessage, um_name="Player.GetGameAchievements"):
    appid: int = betterproto.uint32_field(1)
    language: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GetGameAchievementsResponseAchievement(betterproto.Message):
    internal_name: str = betterproto.string_field(1)
    localized_name: str = betterproto.string_field(2)
    localized_desc: str = betterproto.string_field(3)
    icon: str = betterproto.string_field(4)
    icon_gray: str = betterproto.string_field(5)
    hidden: bool = betterproto.bool_field(6)
    player_percent_unlocked: str = betterproto.string_field(7)


class GetGameAchievementsResponse(UnifiedMessage, um_name="Player.GetGameAchievements"):
    achievements: list[GetGameAchievementsResponseAchievement] = betterproto.message_field(1)


class GetFavoriteBadgeRequest(UnifiedMessage, um_name="Player.GetFavoriteBadge"):
    steamid: int = betterproto.uint64_field(1)


class GetFavoriteBadgeResponse(UnifiedMessage, um_name="Player.GetFavoriteBadge"):
    has_favorite_badge: bool = betterproto.bool_field(1)
    badgeid: int = betterproto.uint32_field(2)
    communityitemid: int = betterproto.uint64_field(3)
    item_type: int = betterproto.uint32_field(4)
    border_color: int = betterproto.uint32_field(5)
    appid: int = betterproto.uint32_field(6)
    level: int = betterproto.uint32_field(7)


class SetFavoriteBadgeRequest(UnifiedMessage, um_name="Player.SetFavoriteBadge"):
    communityitemid: int = betterproto.uint64_field(1)
    badgeid: int = betterproto.uint32_field(2)


class SetFavoriteBadgeResponse(UnifiedMessage, um_name="Player.SetFavoriteBadge"):
    pass


class GetProfileCustomizationRequest(UnifiedMessage, um_name="Player.GetProfileCustomization"):
    steamid: int = betterproto.fixed64_field(1)
    include_inactive_customizations: bool = betterproto.bool_field(2)
    include_purchased_customizations: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class ProfileCustomizationSlot(betterproto.Message):
    slot: int = betterproto.uint32_field(1)
    appid: int = betterproto.uint32_field(2)
    publishedfileid: int = betterproto.uint64_field(3)
    item_assetid: int = betterproto.uint64_field(4)
    item_contextid: int = betterproto.uint64_field(5)
    notes: str = betterproto.string_field(6)
    title: str = betterproto.string_field(7)
    accountid: int = betterproto.uint32_field(8)
    badgeid: int = betterproto.uint32_field(9)
    border_color: int = betterproto.uint32_field(10)
    item_classid: int = betterproto.uint64_field(11)
    item_instanceid: int = betterproto.uint64_field(12)
    ban_check_result: "EBanContentCheckResult" = betterproto.enum_field(13)
    replay_year: int = betterproto.uint32_field(14)


@dataclass(eq=False, repr=False)
class ProfileCustomization(betterproto.Message):
    customization_type: "EProfileCustomizationType" = betterproto.enum_field(1)
    large: bool = betterproto.bool_field(2)
    slots: "list[ProfileCustomizationSlot]" = betterproto.message_field(3)
    active: bool = betterproto.bool_field(4)
    customization_style: "EProfileCustomizationStyle" = betterproto.enum_field(5)
    purchaseid: int = betterproto.uint64_field(6)
    level: int = betterproto.uint32_field(7)


@dataclass(eq=False, repr=False)
class ProfileTheme(betterproto.Message):
    theme_id: str = betterproto.string_field(1)
    title: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ProfilePreferences(betterproto.Message):
    hide_profile_awards: bool = betterproto.bool_field(1)


class GetProfileCustomizationResponse(UnifiedMessage, um_name="Player.GetProfileCustomization"):
    customizations: "list[ProfileCustomization]" = betterproto.message_field(1)
    slots_available: int = betterproto.uint32_field(2)
    profile_theme: "ProfileTheme" = betterproto.message_field(3)
    purchased_customizations: "list[GetProfileCustomizationResponsePurchasedCustomization]" = betterproto.message_field(
        4
    )
    profile_preferences: "ProfilePreferences" = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class GetProfileCustomizationResponsePurchasedCustomization(betterproto.Message):
    purchaseid: int = betterproto.uint64_field(1)
    customization_type: "EProfileCustomizationType" = betterproto.enum_field(2)
    level: int = betterproto.uint32_field(3)


class GetPurchasedProfileCustomizationsRequest(UnifiedMessage, um_name="Player.GetPurchasedProfileCustomizations"):
    steamid: int = betterproto.fixed64_field(1)


class GetPurchasedProfileCustomizationsResponse(UnifiedMessage, um_name="Player.GetPurchasedProfileCustomizations"):
    purchased_customizations: "list[GetPurchasedProfileCustomizationsResponsePurchasedCustomization]" = (
        betterproto.message_field(1)
    )


@dataclass(eq=False, repr=False)
class GetPurchasedProfileCustomizationsResponsePurchasedCustomization(betterproto.Message):
    purchaseid: int = betterproto.uint64_field(1)
    customization_type: "EProfileCustomizationType" = betterproto.enum_field(2)


class GetPurchasedAndUpgradedProfileCustomizationsRequest(
    UnifiedMessage, um_name="Player.GetPurchasedAndUpgradedProfileCustomizations"
):
    steamid: int = betterproto.fixed64_field(1)


class GetPurchasedAndUpgradedProfileCustomizationsResponse(
    UnifiedMessage, um_name="Player.GetPurchasedAndUpgradedProfileCustomizations"
):
    purchased_customizations: "list[GetPurchasedAndUpgradedProfileCustomizationsResponsePurchasedCustomization]" = (
        betterproto.message_field(1)
    )
    upgraded_customizations: "list[GetPurchasedAndUpgradedProfileCustomizationsResponseUpgradedCustomization]" = (
        betterproto.message_field(2)
    )


@dataclass(eq=False, repr=False)
class GetPurchasedAndUpgradedProfileCustomizationsResponsePurchasedCustomization(betterproto.Message):
    customization_type: "EProfileCustomizationType" = betterproto.enum_field(1)
    count: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class GetPurchasedAndUpgradedProfileCustomizationsResponseUpgradedCustomization(betterproto.Message):
    customization_type: "EProfileCustomizationType" = betterproto.enum_field(1)
    level: int = betterproto.uint32_field(2)


class GetProfileThemesAvailableRequest(UnifiedMessage, um_name="Player.GetProfileThemesAvailable"):
    pass


class GetProfileThemesAvailableResponse(UnifiedMessage, um_name="Player.GetProfileThemesAvailable"):
    profile_themes: "list[ProfileTheme]" = betterproto.message_field(1)


class SetProfileThemeRequest(UnifiedMessage, um_name="Player.SetProfileTheme"):
    theme_id: str = betterproto.string_field(1)


class SetProfileThemeResponse(UnifiedMessage, um_name="Player.SetProfileTheme"):
    pass


class SetProfilePreferencesRequest(UnifiedMessage, um_name="Player.SetProfilePreferences"):
    profile_preferences: "ProfilePreferences" = betterproto.message_field(1)


class SetProfilePreferencesResponse(UnifiedMessage, um_name="Player.SetProfilePreferences"):
    pass


class PostStatusToFriendsRequest(UnifiedMessage, um_name="Player.PostStatusToFriends"):
    appid: int = betterproto.uint32_field(1)
    status_text: str = betterproto.string_field(2)


class PostStatusToFriendsResponse(UnifiedMessage, um_name="Player.PostStatusToFriends"):
    pass


class GetPostedStatusRequest(UnifiedMessage, um_name="Player.GetPostedStatus"):
    steamid: int = betterproto.uint64_field(1)
    postid: int = betterproto.uint64_field(2)


class GetPostedStatusResponse(UnifiedMessage, um_name="Player.GetPostedStatus"):
    accountid: int = betterproto.uint32_field(1)
    postid: int = betterproto.uint64_field(2)
    status_text: str = betterproto.string_field(3)
    deleted: bool = betterproto.bool_field(4)
    appid: int = betterproto.uint32_field(5)


class DeletePostedStatusRequest(UnifiedMessage, um_name="Player.DeletePostedStatus"):
    postid: int = betterproto.uint64_field(1)


class DeletePostedStatusResponse(UnifiedMessage, um_name="Player.DeletePostedStatus"):
    pass


class GetLastPlayedTimesRequest(UnifiedMessage, um_name="Player.ClientGetLastPlayedTimes"):
    min_last_played: int = betterproto.uint32_field(1)


class GetLastPlayedTimesResponse(UnifiedMessage, um_name="Player.ClientGetLastPlayedTimes"):
    games: "list[GetLastPlayedTimesResponseGame]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetLastPlayedTimesResponseGame(betterproto.Message):
    appid: int = betterproto.int32_field(1)
    last_playtime: int = betterproto.uint32_field(2)
    playtime_2_weeks: int = betterproto.int32_field(3)
    playtime_forever: int = betterproto.int32_field(4)
    first_playtime: int = betterproto.uint32_field(5)
    playtime_windows_forever: int = betterproto.int32_field(6)
    playtime_mac_forever: int = betterproto.int32_field(7)
    playtime_linux_forever: int = betterproto.int32_field(8)
    first_windows_playtime: int = betterproto.uint32_field(9)
    first_mac_playtime: int = betterproto.uint32_field(10)
    first_linux_playtime: int = betterproto.uint32_field(11)
    last_windows_playtime: int = betterproto.uint32_field(12)
    last_mac_playtime: int = betterproto.uint32_field(13)
    last_linux_playtime: int = betterproto.uint32_field(14)


class AcceptSsaRequest(UnifiedMessage, um_name="Player.AcceptSSA"):
    agreement_type: "EAgreementType" = betterproto.enum_field(1)
    time_signed_utc: int = betterproto.uint32_field(2)


class AcceptSsaResponse(UnifiedMessage, um_name="Player.AcceptSSA"):
    pass


class GetNicknameListRequest(UnifiedMessage, um_name="Player.GetNicknameList"):
    pass


class GetNicknameListResponse(UnifiedMessage, um_name="Player.GetNicknameList"):
    nicknames: "list[GetNicknameListResponsePlayerNickname]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetNicknameListResponsePlayerNickname(betterproto.Message):
    accountid: int = betterproto.fixed32_field(1)
    nickname: str = betterproto.string_field(2)


class GetPerFriendPreferencesRequest(UnifiedMessage, um_name="Player.GetPerFriendPreferences"):
    pass


@dataclass(eq=False, repr=False)
class PerFriendPreferences(betterproto.Message):
    accountid: int = betterproto.fixed32_field(1)
    nickname: str = betterproto.string_field(2)
    notifications_showingame: "ENotificationSetting" = betterproto.enum_field(3)
    notifications_showonline: "ENotificationSetting" = betterproto.enum_field(4)
    notifications_showmessages: "ENotificationSetting" = betterproto.enum_field(5)
    sounds_showingame: "ENotificationSetting" = betterproto.enum_field(6)
    sounds_showonline: "ENotificationSetting" = betterproto.enum_field(7)
    sounds_showmessages: "ENotificationSetting" = betterproto.enum_field(8)
    notifications_sendmobile: "ENotificationSetting" = betterproto.enum_field(9)


class GetPerFriendPreferencesResponse(UnifiedMessage, um_name="Player.GetPerFriendPreferences"):
    preferences: "list[PerFriendPreferences]" = betterproto.message_field(1)


class SetPerFriendPreferencesRequest(UnifiedMessage, um_name="Player.SetPerFriendPreferences"):
    preferences: "PerFriendPreferences" = betterproto.message_field(1)


class SetPerFriendPreferencesResponse(UnifiedMessage, um_name="Player.SetPerFriendPreferences"):
    pass


class AddFriendRequest(UnifiedMessage, um_name="Player.AddFriend"):
    steamid: int = betterproto.fixed64_field(1)


class AddFriendResponse(UnifiedMessage, um_name="Player.AddFriend"):
    invite_sent: bool = betterproto.bool_field(1)
    friend_relationship: int = betterproto.uint32_field(2)
    eresult: int = betterproto.int32_field(3)


class RemoveFriendRequest(UnifiedMessage, um_name="Player.RemoveFriend"):
    steamid: int = betterproto.fixed64_field(1)


class RemoveFriendResponse(UnifiedMessage, um_name="Player.RemoveFriend"):
    friend_relationship: int = betterproto.uint32_field(1)


class IgnoreFriendRequest(UnifiedMessage, um_name="Player.IgnoreFriend"):
    steamid: int = betterproto.fixed64_field(1)
    unignore: bool = betterproto.bool_field(2)


class IgnoreFriendResponse(UnifiedMessage, um_name="Player.IgnoreFriend"):
    friend_relationship: int = betterproto.uint32_field(1)


class GetCommunityPreferencesRequest(UnifiedMessage, um_name="Player.GetCommunityPreferences"):
    pass


@dataclass(eq=False, repr=False)
class CommunityPreferences(betterproto.Message):
    hide_adult_content_violence: bool = betterproto.bool_field(1)
    hide_adult_content_sex: bool = betterproto.bool_field(2)
    parenthesize_nicknames: bool = betterproto.bool_field(4)
    text_filter_setting: "ETextFilterSetting" = betterproto.enum_field(5)
    text_filter_ignore_friends: bool = betterproto.bool_field(6)
    text_filter_words_revision: int = betterproto.uint32_field(7)
    timestamp_updated: int = betterproto.uint32_field(3)


class GetCommunityPreferencesResponse(UnifiedMessage, um_name="Player.GetCommunityPreferences"):
    preferences: "CommunityPreferences" = betterproto.message_field(1)


class SetCommunityPreferencesRequest(UnifiedMessage, um_name="Player.SetCommunityPreferences"):
    preferences: "CommunityPreferences" = betterproto.message_field(1)


class SetCommunityPreferencesResponse(UnifiedMessage, um_name="Player.SetCommunityPreferences"):
    pass


class GetTextFilterWordsRequest(UnifiedMessage, um_name="Player.GetTextFilterWords"):
    pass


@dataclass(eq=False, repr=False)
class TextFilterWords(betterproto.Message):
    text_filter_custom_banned_words: list[str] = betterproto.string_field(1)
    text_filter_custom_clean_words: list[str] = betterproto.string_field(2)
    text_filter_words_revision: int = betterproto.uint32_field(3)


class GetTextFilterWordsResponse(UnifiedMessage, um_name="Player.GetTextFilterWords"):
    words: "TextFilterWords" = betterproto.message_field(1)


class GetNewSteamAnnouncementStateRequest(UnifiedMessage, um_name="Player.GetNewSteamAnnouncementState"):
    language: int = betterproto.int32_field(1)


class GetNewSteamAnnouncementStateResponse(UnifiedMessage, um_name="Player.GetNewSteamAnnouncementState"):
    state: int = betterproto.int32_field(1)
    announcement_headline: str = betterproto.string_field(2)
    announcement_url: str = betterproto.string_field(3)
    time_posted: int = betterproto.uint32_field(4)
    announcement_gid: int = betterproto.uint64_field(5)


class UpdateSteamAnnouncementLastReadRequest(UnifiedMessage, um_name="Player.UpdateSteamAnnouncementLastRead"):
    announcement_gid: int = betterproto.uint64_field(1)
    time_posted: int = betterproto.uint32_field(2)


class UpdateSteamAnnouncementLastReadResponse(UnifiedMessage, um_name="Player.UpdateSteamAnnouncementLastRead"):
    pass


class GetPrivacySettingsRequest(UnifiedMessage, um_name="Player.GetPrivacySettings"):
    pass


@dataclass(eq=False, repr=False)
class CPrivacySettings(betterproto.Message):
    privacy_state: int = betterproto.int32_field(1)
    privacy_state_inventory: int = betterproto.int32_field(2)
    privacy_state_gifts: int = betterproto.int32_field(3)
    privacy_state_ownedgames: int = betterproto.int32_field(4)
    privacy_state_playtime: int = betterproto.int32_field(5)
    privacy_state_friendslist: int = betterproto.int32_field(6)


class GetPrivacySettingsResponse(UnifiedMessage, um_name="Player.GetPrivacySettings"):
    privacy_settings: "CPrivacySettings" = betterproto.message_field(1)


class GetDurationControlRequest(UnifiedMessage, um_name="Player.GetDurationControl"):
    appid: int = betterproto.uint32_field(1)


class GetDurationControlResponse(UnifiedMessage, um_name="Player.GetDurationControl"):
    is_enabled: bool = betterproto.bool_field(1)
    seconds: int = betterproto.int32_field(2)
    seconds_today: int = betterproto.int32_field(3)
    is_steamchina_account: bool = betterproto.bool_field(4)
    is_age_verified: bool = betterproto.bool_field(5)
    seconds_allowed_today: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class LastPlayedTimesNotification(betterproto.Message):
    games: "list[GetLastPlayedTimesResponseGame]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class FriendNicknameChangedNotification(betterproto.Message):
    accountid: int = betterproto.fixed32_field(1)
    nickname: str = betterproto.string_field(2)
    is_echo_to_self: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class FriendEquippedProfileItemsChangedNotification(betterproto.Message):
    accountid: int = betterproto.fixed32_field(1)


@dataclass(eq=False, repr=False)
class NewSteamAnnouncementStateNotification(betterproto.Message):
    state: int = betterproto.int32_field(1)
    announcement_headline: str = betterproto.string_field(2)
    announcement_url: str = betterproto.string_field(3)
    time_posted: int = betterproto.uint32_field(4)
    announcement_gid: int = betterproto.uint64_field(5)


@dataclass(eq=False, repr=False)
class CommunityPreferencesChangedNotification(betterproto.Message):
    preferences: "CommunityPreferences" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class TextFilterWordsChangedNotification(betterproto.Message):
    words: "TextFilterWords" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class PerFriendPreferencesChangedNotification(betterproto.Message):
    accountid: int = betterproto.fixed32_field(1)
    preferences: "PerFriendPreferences" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class PrivacySettingsChangedNotification(betterproto.Message):
    privacy_settings: "CPrivacySettings" = betterproto.message_field(1)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from typing import TypedDict

from .id import ID32, AppID, BundleID, PackageID


class Bundle(TypedDict):
    bundleid: BundleID
    name: str
    packageids: list[PackageID]
    appids: list[AppID]
    header_image_url: str
    main_capsule: str
    initial_price: int
    final_price: int
    formatted_orig_price: str
    formatted_final_price: str
    discount_percent: int
    available_windows: bool
    available_mac: bool
    available_linux: bool
    support_vrhmod: bool
    support_vrhmod_only: bool
    creator_clan_ids: list[ID32]
    localized_langs: list[int]
    coming_soon: bool
    library_Asset: str
    no_main_cap: bool
    deck_compatibility_category: int



# pyright: reportUnnecessaryTypeIgnoreComment=false
"""
Various constants/types for use around the library.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE
"""

from __future__ import annotations

import builtins
import importlib.util
import struct
from collections.abc import Callable
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import partial
from io import BytesIO, StringIO
from typing import TYPE_CHECKING, Any, Final, Literal, Protocol, TypeAlias, TypeVar, cast, final

from multidict import MultiDict
from yarl import URL as _URL

from .types.vdf import BinaryVDFDict, VDFDict

if TYPE_CHECKING:
    from typing_extensions import Buffer

    from .clan import Clan
    from .group import Group
    from .state import ConnectionState

DOCS_BUILDING: bool = getattr(builtins, "__sphinx__", False)

HAS_ORJSON = False
try:
    import orjson  # type: ignore
except ModuleNotFoundError:
    import json

    json_loads = json.loads

    @partial(cast, Callable[[Any], str])
    def json_dumps(
        obj: Any,
        __func: Callable[..., str] = json.dumps,
        /,
    ) -> str:
        return __func(obj, separators=(",", ":"), ensure_ascii=True)

else:
    json_loads = orjson.loads

    @partial(cast, Callable[[Any], str])
    def json_dumps(
        obj: Any,
        __func: Callable[[Any], bytes] = orjson.dumps,  # type: ignore
        __decoder: Callable[[bytes], str] = bytes.decode,
        /,
    ) -> str:
        return __decoder(__func(obj))


JSON_LOADS: Final = cast(Callable[[str | bytes], Any], json_loads)
JSON_DUMPS: Final = json_dumps


try:
    import orvdf  # type: ignore
except ModuleNotFoundError:
    import vdf

    def _multi_dict_ify(
        x: Any,
        __isinstance=isinstance,  # type: ignore
        __vdf_dict=vdf.VDFDict,
        __multi_dict: type[MultiDict[Any]] = MultiDict,
        /,
    ):
        if __isinstance(x, __vdf_dict):
            multi_dict = __multi_dict()
            adder = multi_dict.add
            for k, v in x.items():
                adder(k, _multi_dict_ify(v))
            return multi_dict
        return x

    def vdf_loads(
        s: str,
        __func: Callable[..., Any] = vdf.parse,  # type: ignore
        __mapper: type[vdf.VDFDict] = vdf.VDFDict,
        __multi_dict_ify: Callable[[vdf.VDFDict], MultiDict[Any]] = _multi_dict_ify,
        __string_io: type[StringIO] = StringIO,
        /,
    ) -> VDFDict:
        return __multi_dict_ify(__func(__string_io(s), mapper=__mapper))

    def vdf_binary_loads(
        s: bytes,
        __func: Callable[..., Any] = vdf.binary_load,  # type: ignore
        __mapper: type[vdf.VDFDict] = vdf.VDFDict,
        __multi_dict_ify: Callable[[vdf.VDFDict], MultiDict[Any]] = _multi_dict_ify,
        __bytes_io: type[BytesIO] = BytesIO,
        /,
    ) -> BinaryVDFDict:
        return __multi_dict_ify(__func(__bytes_io(s), mapper=__mapper))

else:
    vdf_loads = orvdf.loads  # type: ignore
    vdf_binary_loads = orvdf.binary_loads  # type: ignore

VDF_LOADS: Final = cast(Callable[[str], VDFDict], vdf_loads)
VDF_BINARY_LOADS: Final = cast(Callable[[bytes], BinaryVDFDict], vdf_binary_loads)

HTML_PARSER: Final = "lxml-xml" if importlib.util.find_spec("lxml") else "html.parser"

try:
    from asyncio import TaskGroup as TaskGroup, timeout as timeout  # type: ignore
except ImportError:
    from taskgroup import TaskGroup as TaskGroup, timeout as timeout  # type: ignore


UNIX_EPOCH: Final = datetime(1970, 1, 1, tzinfo=timezone.utc)
STEAM_EPOCH: Final = datetime(2003, 1, 1, tzinfo=timezone.utc)


def READ_U32(
    s: Buffer, unpacker: Callable[[Buffer], tuple[int]] = cast(Any, struct.Struct("<I").unpack_from), /
) -> int:
    (u32,) = unpacker(s)
    return u32


WRITE_U32: Final = cast(Callable[[int], bytes], struct.Struct("<I").pack)

_PROTOBUF_MASK: Final = 0x80000000
# inlined as these are some of the most called functions in the library
IS_PROTO: Final = cast(Callable[[int], bool], _PROTOBUF_MASK.__and__)  # this is boolean like for a bit of extra speed
SET_PROTO_BIT: Final = _PROTOBUF_MASK.__or__
CLEAR_PROTO_BIT: Final = (~_PROTOBUF_MASK).__and__

STATE = ContextVar["ConnectionState"]("APP_STATE")


@final
class MissingSentinel(Any if TYPE_CHECKING else object):
    __slots__ = ()

    def __eq__(self, other: object) -> Literal[False]:
        return False

    def __bool__(self) -> Literal[False]:
        return False

    def __hash__(self) -> Literal[0]:
        return 0

    def __repr__(self) -> Literal["..."]:
        return "..."


MISSING: Final = MissingSentinel()


@final
class URL:
    API: Final = _URL("https://api.steampowered.com")
    COMMUNITY: Final = _URL("https://steamcommunity.com")
    STORE: Final = _URL("https://store.steampowered.com")
    HELP: Final = _URL("https://help.steampowered.com")
    LOGIN: Final = _URL("https://login.steampowered.com")
    CDN: Final = _URL("https://cdn.cloudflare.steamstatic.com")
    PUBLISHER_API: Final = _URL("https://partner.steam-api.com")


DEFAULT_AVATAR: Final = b"\xfe\xf4\x9e\x7f\xa7\xe1\x99s\x10\xd7\x05\xb2\xa6\x15\x8f\xf8\xdc\x1c\xdf\xeb"


class _IDComparable(Protocol):
    @property
    def id(self) -> Any: ...


_TT_IDComp = TypeVar("_TT_IDComp", bound=type[_IDComparable])


def impl_eq_via_id(cls: _TT_IDComp) -> _TT_IDComp:
    def __eq__(self: _IDComparable, other: object, /) -> bool:
        return isinstance(other, cls) and self.id == other.id  # type: ignore

    def __hash__(self: _IDComparable) -> int:
        return hash(self.id)

    cls.__eq__ = __eq__
    cls.__hash__ = __hash__
    return cls


class _HasChatGroupMixin(Protocol):
    __slots__ = ()
    clan: Clan | None
    group: Group | None

    @property
    def _chat_group(self) -> Clan | Group:
        chat_group = self.clan or self.group
        assert chat_group is not None
        return chat_group


T_co = TypeVar("T_co", covariant=True)


class _ReadOnlyProto(Protocol[T_co]):
    def __get__(self, __instance: Any, __owner: type) -> T_co: ...


ReadOnly: TypeAlias = T_co | _ReadOnlyProto[T_co]
"""PEP 705 style read only, should make things a easier transition when the feature gets added to nominal classes.

Currently does not mean anything to a type checker really.
"""


@dataclass(frozen=True, slots=True)
class SteamBadge:
    name: str
    id: int
    url: str
    level: float = 1


STEAM_BADGES = (
    SteamBadge(
        "Years of Service",
        1,
        "https://community.cloudflare.steamstatic.com/public/images/badges/02_years/steamyears20_80.png",
        20,
    ),
    SteamBadge(
        "Community Ambassador",
        2,
        "https://community.cloudflare.steamstatic.com/public/images/badges/01_community/community03_80.png",
        3,
    ),
    SteamBadge(
        "The Potato Sack",
        3,
        "https://community.cloudflare.steamstatic.com/public/images/badges/03_potato/potato03_80.png",
        36,
    ),
    SteamBadge(
        "The Great Steam Treasure Hunt",
        4,
        "https://community.cloudflare.steamstatic.com/public/images/badges/04_treasurehunt/treasurehunt03_80.png",
        3,
    ),
    SteamBadge(
        "Steam Summer Camp",
        5,
        "https://community.cloudflare.steamstatic.com/public/images/badges/05_summer2011/tickets80.png",
        76,
    ),
    SteamBadge(
        "Steam Holiday Sale 2011",
        6,
        "https://community.cloudflare.steamstatic.com/public/images/badges/06_winter2011/coal03_80.png",
        78,
    ),
    SteamBadge(
        "Steam Summer Sale 2012",
        7,
        "https://community.cloudflare.steamstatic.com/public/images/badges/07_summer2012/Summer2012_stage3_80.png",
        3,
    ),
    SteamBadge(
        "Steam Holiday Sale 2012",
        8,
        "https://community.cloudflare.steamstatic.com/public/images/badges/08_winter2012/winter2012_badge80.png",
    ),
    SteamBadge(
        "Steam Community Translator",
        9,
        "https://community.cloudflare.steamstatic.com/public/images/badges/09_communitytranslator/translator_level4_80.png",
        4,
    ),
    SteamBadge(
        "Steam Community Moderator",
        10,
        "https://community.cloudflare.steamstatic.com/public/images/badges/generic/CommunityModerator_80.png",
    ),
    SteamBadge(
        "Valve Employee",
        11,
        "https://community.cloudflare.steamstatic.com/public/images/badges/generic/ValveEmployee_80.png",
    ),
    SteamBadge(
        "Steamworks Developer",
        12,
        "https://community.cloudflare.steamstatic.com/public/images/badges/generic/GameDeveloper_80.png",
    ),
    SteamBadge(
        "Owned Games",
        13,
        "https://community.cloudflare.steamstatic.com/public/images/badges/13_gamecollector/30000_80.png?v=4",
        float("inf"),
    ),
    SteamBadge(
        "Trading Card Beta Tester",
        14,
        "https://community.cloudflare.steamstatic.com/public/images/badges/generic/TradingCardBeta_80.png",
    ),
    SteamBadge(
        "Steam Hardware Beta",
        15,
        "https://community.cloudflare.steamstatic.com/public/images/badges/15_hwbeta/hwbeta03_80.png",
        3,
    ),
    SteamBadge(
        "Red Team",
        16,
        "https://community.cloudflare.steamstatic.com/public/images/badges/16_summer2014/team_red.png",
        float("inf"),
    ),
    SteamBadge(
        "Blue Team",
        17,
        "https://community.cloudflare.steamstatic.com/public/images/badges/16_summer2014/team_blue.png",
        float("inf"),
    ),
    SteamBadge(
        "Pink Team",
        18,
        "https://community.cloudflare.steamstatic.com/public/images/badges/16_summer2014/team_pink.png",
        float("inf"),
    ),
    SteamBadge(
        "Green Team",
        19,
        "https://community.cloudflare.steamstatic.com/public/images/badges/16_summer2014/team_green.png",
        float("inf"),
    ),
    SteamBadge(
        "Purple Team",
        20,
        "https://community.cloudflare.steamstatic.com/public/images/badges/16_summer2014/team_purple.png",
        float("inf"),
    ),
    SteamBadge(
        "Auction Participant/Winner",
        21,
        "https://community.cloudflare.steamstatic.com/public/images/badges/21_auction/winner_80.png?v=2",
        3,
    ),
    SteamBadge(
        "2014 Holiday Profile Recipient",
        22,
        "https://community.cloudflare.steamstatic.com/public/images/badges/22_golden/owner_80.png",
    ),
    SteamBadge(
        "Monster Summer",
        23,
        "https://community.cloudflare.steamstatic.com/public/images/badges/23_towerattack/wormhole.png",
        100_000_000,
    ),
    SteamBadge(
        "Red Herring",
        24,
        "https://community.cloudflare.steamstatic.com/public/images/badges/24_winter2015_arg_red_herring/red_herring.png",
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2016",
        25,
        "https://community.cloudflare.steamstatic.com/public/images/badges/25_steamawardnominations/level04_80.png",
        4,
    ),
    SteamBadge(
        "Sticker Completionist",
        26,
        "https://community.cloudflare.steamstatic.com/public/images/badges/26_summer2017_sticker/completionist.png",
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2017",
        27,
        "https://community.cloudflare.steamstatic.com/public/images/badges/27_steamawardnominations/level04_80.png",
        4,
    ),
    SteamBadge(
        "Spring Cleaning Event 2018",
        28,
        "https://community.cloudflare.steamstatic.com/public/images/badges/28_springcleaning2018/gold_80.png",
        3,
    ),
    SteamBadge(
        "Salien", 29, "https://community.cloudflare.steamstatic.com/public/images/badges/29_salien/6_80.png", 29
    ),
    SteamBadge(
        "Retired Community Moderator",
        30,
        "https://community.cloudflare.steamstatic.com/public/images/badges/generic/RetiredModerator_80.png",
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2018",
        31,
        "https://community.cloudflare.steamstatic.com/public/images/badges/30_steamawardnominations/level04_80.png",
        4,
    ),
    SteamBadge(
        "Winter 2018 Knick-Knack Collector",
        33,
        "https://community.cloudflare.steamstatic.com/public/images/badges/33_cozycottage2018/1000000_80.png",
        float("inf"),
    ),
    SteamBadge(
        "Lunar New Year 2019",
        34,
        "https://community.cloudflare.steamstatic.com/public/images/badges/34_lny2019/10_80.png",
        10,
    ),
    SteamBadge(
        "Spring Cleaning Event 2019",
        36,
        "https://community.cloudflare.steamstatic.com/public/images/badges/36_springcleaning2019/gold_80x80.png",
        3,
    ),
    SteamBadge(
        "Steam Grand Prix 2019",
        37,
        "https://community.cloudflare.steamstatic.com/public/images/badges/37_summer2019/level1000000_80.png",
        float("inf"),
    ),
    SteamBadge(
        "Team Hare",
        38,
        "https://community.cloudflare.steamstatic.com/public/images/badges/37_summer2019/hare_gold_80.png",
    ),
    SteamBadge(
        "Team Tortoise",
        39,
        "https://community.cloudflare.steamstatic.com/public/images/badges/37_summer2019/tortoise_gold_80.png",
    ),
    SteamBadge(
        "Team Corgi",
        40,
        "https://community.cloudflare.steamstatic.com/public/images/badges/37_summer2019/corgi_gold_80.png",
    ),
    SteamBadge(
        "Team Cockatiel",
        41,
        "https://community.cloudflare.steamstatic.com/public/images/badges/37_summer2019/cockatiel_gold_80.png",
    ),
    SteamBadge(
        "Team Pig",
        42,
        "https://community.cloudflare.steamstatic.com/public/images/badges/37_summer2019/pig_gold_80.png",
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2019",
        43,
        "https://community.cloudflare.steamstatic.com/public/images/badges/43_steamawardnominations/level04_80.png",
        4,
    ),
    SteamBadge(
        "Steam Winter Sale 2019",
        44,
        "https://community.cloudflare.steamstatic.com/public/images/badges/44_winter2019/level15_80.png",
        15,
    ),
    SteamBadge(
        "Steamville 2019 Badge",
        45,
        "https://community.cloudflare.steamstatic.com/public/images/badges/45_steamville2019/key_to_city_80.png",
    ),
    SteamBadge(
        "Lunar New Year 2020",
        46,
        "https://community.cloudflare.steamstatic.com/public/images/badges/46_lny2020/10_80.png",
        10,
    ),
    SteamBadge(
        "Spring Cleaning Event 2020",
        47,
        "https://community.cloudflare.steamstatic.com/public/images/badges/47_springcleaning2020/dewey_badge_3.0_80x80.png",
        3,
    ),
    SteamBadge(
        "Community Contributor",
        48,
        "https://community.cloudflare.steamstatic.com/public/images/badges/48_communitycontributor/1_80.png",
        float("inf"),
    ),
    SteamBadge(
        "Community Patron",
        49,
        "https://community.cloudflare.steamstatic.com/public/images/badges/49_communitypatron/1_80.png",
        float("inf"),
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2020",
        50,
        "https://community.cloudflare.steamstatic.com/public/images/badges/50_steamawardnominations/level04_184.png",
        4,
    ),
    SteamBadge(
        "The Masked Avenger",
        51,
        "https://community.cloudflare.steamstatic.com/public/images/badges/51_summer2021/51.png",
    ),
    SteamBadge(
        "The Trailblazing Explorer",
        52,
        "https://community.cloudflare.steamstatic.com/public/images/badges/51_summer2021/52.png",
    ),
    SteamBadge(
        "The Gorilla Scientist",
        53,
        "https://community.cloudflare.steamstatic.com/public/images/badges/51_summer2021/53.png",
    ),
    SteamBadge(
        "The Paranormal Professor",
        54,
        "https://community.cloudflare.steamstatic.com/public/images/badges/51_summer2021/54.png",
    ),
    SteamBadge(
        "The Ghost Detective",
        55,
        "https://community.cloudflare.steamstatic.com/public/images/badges/51_summer2021/55.png",
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2021",
        56,
        "https://community.cloudflare.steamstatic.com/public/images/badges/56_steamawardnominations/level04_80.png",
        4,
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2021 Classic Edition",
        57,
        "https://community.cloudflare.steamstatic.com/public/images/badges/57_steamawardnominationsclassic/2021_nomination_classic_level04.png",
        4,
    ),
    SteamBadge(
        "2022 Steam Cup",
        59,
        "https://community.cloudflare.steamstatic.com/public/images/badges/generic/RacingSale2022_80.png",
    ),
    SteamBadge(
        "2022 Steam Next Fest June Edition",
        60,
        "https://community.cloudflare.steamstatic.com/public/images/badges/60_nextfestsummer2022/160/badge_3.png",
        1056,
    ),
    SteamBadge(
        "Clorthax's Paradox Party Badge",
        61,
        "https://community.cloudflare.steamstatic.com/public/images/badges/61_summersale_minigame2022/80/badge_clorthax_3.png?t=7337941",
        10,
    ),
    SteamBadge(
        "2022 Steam Next Fest October Edition",
        62,
        "https://community.cloudflare.steamstatic.com/public/images/badges/62_nextfestautumn2022/160/badge_3.png",
        6,
    ),
    SteamBadge(
        "Steam Awards Nomination Committee 2022",
        63,
        "https://community.cloudflare.steamstatic.com/public/images/badges/63_steamawardnominations/level04_80.png",
        4,
    ),
    SteamBadge(
        "Steam Replay 2022",
        64,
        "https://community.akamai.steamstatic.com/public/images/badges/generic/Replay2022_80.png",
    ),
)


# default CMs if Steam API is down
DEFAULT_CMS: Final = (
    "cmp1-ams1.steamserver.net:27018",
    "cmp1-ams1.steamserver.net:27018",
    "cmp1-ams1.steamserver.net:27018",
    "cmp1-ams1.steamserver.net:27018",
    "cmp1-ams1.steamserver.net:27018",
    "cmp1-ams1.steamserver.net:443",
    "cmp1-ams1.steamserver.net:443",
    "cmp1-ams1.steamserver.net:443",
    "cmp1-ams1.steamserver.net:443",
    "cmp1-ams1.steamserver.net:443",
    "cmp1-ams1.steamserver.net:443",
    "cmp1-atl3.steamserver.net:27018",
    "cmp1-atl3.steamserver.net:27018",
    "cmp1-atl3.steamserver.net:27018",
    "cmp1-atl3.steamserver.net:27018",
    "cmp1-atl3.steamserver.net:443",
    "cmp1-atl3.steamserver.net:443",
    "cmp1-atl3.steamserver.net:443",
    "cmp1-fra1.steamserver.net:27019",
    "cmp1-fra1.steamserver.net:27019",
    "cmp1-fra1.steamserver.net:27020",
    "cmp1-fra1.steamserver.net:27020",
    "cmp1-fra1.steamserver.net:27022",
    "cmp1-fra1.steamserver.net:27024",
    "cmp1-fra1.steamserver.net:27024",
    "cmp1-fra2.steamserver.net:27018",
    "cmp1-fra2.steamserver.net:27019",
    "cmp1-fra2.steamserver.net:27019",
    "cmp1-fra2.steamserver.net:27019",
    "cmp1-fra2.steamserver.net:27019",
    "cmp1-fra2.steamserver.net:27019",
    "cmp1-fra2.steamserver.net:27021",
    "cmp1-fra2.steamserver.net:27023",
    "cmp1-fra2.steamserver.net:27023",
    "cmp1-fra2.steamserver.net:27024",
    "cmp1-fra2.steamserver.net:443",
    "cmp1-iad1.steamserver.net:27018",
    "cmp1-iad1.steamserver.net:27018",
    "cmp1-iad1.steamserver.net:27018",
    "cmp1-iad1.steamserver.net:27018",
    "cmp1-iad1.steamserver.net:27018",
    "cmp1-iad1.steamserver.net:27019",
    "cmp1-iad1.steamserver.net:27019",
    "cmp1-iad1.steamserver.net:27019",
    "cmp1-iad1.steamserver.net:27019",
    "cmp1-iad1.steamserver.net:27020",
    "cmp1-iad1.steamserver.net:27020",
    "cmp1-iad1.steamserver.net:27020",
    "cmp1-iad1.steamserver.net:27020",
    "cmp1-iad1.steamserver.net:27020",
    "cmp1-iad1.steamserver.net:443",
    "cmp1-iad1.steamserver.net:443",
    "cmp1-lhr1.steamserver.net:27018",
    "cmp1-lhr1.steamserver.net:27018",
    "cmp1-lhr1.steamserver.net:27018",
    "cmp1-lhr1.steamserver.net:27018",
    "cmp1-lhr1.steamserver.net:27018",
    "cmp1-lhr1.steamserver.net:27019",
    "cmp1-lhr1.steamserver.net:443",
    "cmp1-lhr1.steamserver.net:443",
    "cmp1-lhr1.steamserver.net:443",
    "cmp1-lhr1.steamserver.net:443",
    "cmp1-lhr1.steamserver.net:443",
    "cmp1-lhr1.steamserver.net:443",
    "cmp1-lhr1.steamserver.net:443",
    "cmp1-ord1.steamserver.net:27018",
    "cmp1-ord1.steamserver.net:27018",
    "cmp2-ams1.steamserver.net:27018",
    "cmp2-ams1.steamserver.net:27018",
    "cmp2-ams1.steamserver.net:27018",
    "cmp2-ams1.steamserver.net:27018",
    "cmp2-ams1.steamserver.net:27018",
    "cmp2-ams1.steamserver.net:27018",
    "cmp2-ams1.steamserver.net:443",
    "cmp2-ams1.steamserver.net:443",
    "cmp2-ams1.steamserver.net:443",
    "cmp2-ams1.steamserver.net:443",
    "cmp2-atl3.steamserver.net:27018",
    "cmp2-atl3.steamserver.net:27018",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-atl3.steamserver.net:443",
    "cmp2-fra1.steamserver.net:27019",
    "cmp2-fra1.steamserver.net:27019",
    "cmp2-fra1.steamserver.net:27019",
    "cmp2-fra1.steamserver.net:27019",
    "cmp2-fra1.steamserver.net:27020",
    "cmp2-fra1.steamserver.net:27020",
    "cmp2-fra1.steamserver.net:27020",
    "cmp2-fra1.steamserver.net:27020",
    "cmp2-fra1.steamserver.net:27022",
    "cmp2-fra1.steamserver.net:27022",
    "cmp2-fra1.steamserver.net:27023",
    "cmp2-fra1.steamserver.net:27023",
    "cmp2-fra1.steamserver.net:27024",
    "cmp2-fra1.steamserver.net:27024",
    "cmp2-fra2.steamserver.net:27019",
    "cmp2-fra2.steamserver.net:27019",
    "cmp2-fra2.steamserver.net:27019",
    "cmp2-fra2.steamserver.net:27020",
    "cmp2-fra2.steamserver.net:27021",
    "cmp2-fra2.steamserver.net:27022",
    "cmp2-fra2.steamserver.net:27022",
    "cmp2-fra2.steamserver.net:27023",
    "cmp2-fra2.steamserver.net:27024",
    "cmp2-fra2.steamserver.net:27024",
    "cmp2-fra2.steamserver.net:27024",
    "cmp2-fra2.steamserver.net:27024",
    "cmp2-fra2.steamserver.net:27024",
    "cmp2-iad1.steamserver.net:27018",
    "cmp2-iad1.steamserver.net:27019",
    "cmp2-iad1.steamserver.net:27019",
    "cmp2-iad1.steamserver.net:27019",
    "cmp2-iad1.steamserver.net:27020",
    "cmp2-lhr1.steamserver.net:27018",
    "cmp2-lhr1.steamserver.net:27018",
    "cmp2-lhr1.steamserver.net:27019",
    "cmp2-lhr1.steamserver.net:27020",
    "cmp2-lhr1.steamserver.net:443",
    "cmp2-lhr1.steamserver.net:443",
    "cmp2-lhr1.steamserver.net:443",
    "cmp2-lhr1.steamserver.net:443",
    "cmp2-lhr1.steamserver.net:443",
    "cmp2-ord1.steamserver.net:27018",
    "cmp2-ord1.steamserver.net:27018",
    "cmp2-ord1.steamserver.net:27018",
    "cmp2-ord1.steamserver.net:27018",
    "cmp2-ord1.steamserver.net:27018",
    "cmp2-ord1.steamserver.net:27018",
    "cmp2-ord1.steamserver.net:27018",
    "cmp2-ord1.steamserver.net:27018",
    "ext1-par1.steamserver.net:27019",
    "ext1-par1.steamserver.net:27021",
    "ext1-par1.steamserver.net:27024",
    "ext1-par1.steamserver.net:27028",
    "ext1-par1.steamserver.net:27032",
    "ext1-par1.steamserver.net:27033",
    "ext1-par1.steamserver.net:27034",
    "ext1-par1.steamserver.net:27037",
    "ext1-par1.steamserver.net:443",
    "ext2-par1.steamserver.net:27021",
    "ext2-par1.steamserver.net:27022",
    "ext2-par1.steamserver.net:27023",
    "ext2-par1.steamserver.net:27024",
    "ext2-par1.steamserver.net:27025",
    "ext2-par1.steamserver.net:27028",
    "ext2-par1.steamserver.net:27030",
    "ext2-par1.steamserver.net:27035",
    "ext2-par1.steamserver.net:27036",
    "ext2-par1.steamserver.net:27038",
    "ext2-par1.steamserver.net:443",
)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import abc
import asyncio
import sys
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Generic, Literal, Protocol, TypedDict, cast, runtime_checkable

from bs4 import BeautifulSoup
from typing_extensions import Required, Self, TypeVar
from yarl import URL as URL_

from . import utils
from ._const import HTML_PARSER, MISSING, STEAM_EPOCH, UNIX_EPOCH, URL, VDF_BINARY_LOADS, ReadOnly
from .achievement import UserAppAchievement, UserAppStats
from .app import *
from .badge import FavouriteBadge, UserBadges
from .enums import *
from .errors import ConfirmationError, HTTPException, WSException
from .id import ID
from .models import Avatar, Ban
from .profile import *
from .protobufs import econ
from .reaction import Award, AwardReaction, Emoticon, MessageReaction, PartialMessageReaction, Sticker
from .trade import Asset, Inventory, Item, TradeOffer
from .types.id import ID64, AppID, AssetID, CommentID, ContextID, Intable, PostID, PublishedFileID
from .types.user import UserT
from .utils import DateTime, classproperty, parse_bb_code

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Coroutine, Sequence
    from ipaddress import IPv4Address

    from .clan import Clan
    from .comment import Comment
    from .game_server import GameServer
    from .group import Group
    from .invite import AppInvite
    from .media import Media
    from .message import UserMessage
    from .post import Post
    from .published_file import PublishedFile
    from .review import Review
    from .state import ConnectionState
    from .types import achievement
    from .user import ClientUser, User

__all__ = (
    "Message",
    "Channel",
    "PartialUser",
    "UserInventoryInfo",
)

MessageT = TypeVar("MessageT", bound="Message", default="Message", covariant=True)


class _CommentableKwargs(TypedDict, total=False):
    id64: Required[int]
    topic_id: int
    forum_id: int


class _CommentThreadType(IntEnum):
    # from https://github.com/SteamDatabase/Protobufs/blob/62702bd83148fff64945f20a09eaf8f250782829/steam/enums.proto#L386
    WorkshopAccountDeveloper = 2
    WorkshopAccount = 3
    PublishedFileDeveloper = 4
    PublishedFile = 5
    Test = 6
    Topic = 7
    Review = 8
    User = 10
    NewsPost = 11
    Clan = 12
    Announcement = 13
    Event = 14
    Post = 15
    UserReceivedNewGame = 16
    PublishedFileAnnouncement = 17
    ModeratorMessage = 18
    ClanCuratedApp = 19
    QAndASession = 20


class Commentable(Protocol):
    """A mixin that implements commenting functionality."""

    __slots__ = ()
    _state: ConnectionState

    @property
    @abc.abstractmethod
    def _commentable_kwargs(self) -> _CommentableKwargs:
        raise NotImplementedError

    @classproperty
    def _COMMENTABLE_TYPE(cls: type[Self]) -> _CommentThreadType:
        try:
            return _CommentThreadType[cls.__name__]
        except KeyError:
            raise AttributeError(f"{cls.__name__} is not a valid commentable type")

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if (
            cls.__dict__.get("_COMMENTABLE_TYPE") is Commentable.__dict__["_COMMENTABLE_TYPE"]
            and cls.__name__ != "BaseEvent"
        ):
            assert cls.__name__ in _CommentThreadType.__members__, f"{cls.__name__} is not a valid commentable type"

    async def fetch_comment(self, id: int, /) -> Comment[Self]:
        """Fetch a comment by its ID.

        Parameters
        ----------
        id
            The ID of the comment to fetch.
        """
        from .comment import Comment

        comment = await self._state.fetch_comment(self, id)
        return Comment(
            self._state,
            id=CommentID(comment.id),
            content=comment.content,
            created_at=DateTime.from_timestamp(comment.timestamp),
            author=await self._state._maybe_user(comment.author_id64),
            owner=self,
            reactions=[AwardReaction(self._state, reaction) for reaction in comment.reactions],
        )

    async def comment(self, content: str, /, *, subscribe: bool = True) -> Comment[Self, ClientUser]:
        """Post a comment to a comments section.

        Parameters
        ----------
        content
            The message to add to the comment section.
        subscribe
            Whether to subscribe to notifications on any future activity in this comment's thread.

        Returns
        -------
        The created comment.
        """
        return await self._state.post_comment(self, content, subscribe)

    async def comments(
        self,
        *,
        oldest_first: bool = False,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[Comment[Self], None]:
        """An :term:`asynchronous iterator` for accessing a comment section's :class:`~steam.Comment` objects.

        Examples
        --------

        Usage:

        .. code:: python3

            async for comment in commentable.comments(limit=10):
                print("Author:", comment.author, "Said:", comment.content)

        All parameters are optional.

        Parameters
        ----------
        oldest_first
            Whether or not to request comments with the oldest comments first or last. Defaults to ``False``.
        limit
            The maximum number of comments to search through.
            Default is ``None`` which will fetch the all the comments in the comments section.
        before
            A time to search for comments before.
        after
            A time to search for comments after.

        Yields
        ---------
        :class:`~steam.Comment`
        """
        from .comment import Comment
        from .reaction import AwardReaction

        after = after or UNIX_EPOCH
        before = before or DateTime.now()
        count: int | None = None
        total_count = 0
        yielded = 0

        async def get_comments(chunk: int) -> list[Comment[Self]]:
            nonlocal after, before, count, total_count, yielded

            starting_from = total_count - count if total_count and count else 0
            proto = await self._state.fetch_comments(self, chunk, starting_from, oldest_first)

            if count is None:
                total_count = count = proto.total_count

            comments: list[Comment[Self]] = []
            comment = None
            for comment in proto.comments:
                comment = Comment(
                    self._state,
                    id=CommentID(comment.id),
                    content=comment.content,
                    created_at=DateTime.from_timestamp(comment.timestamp),
                    reactions=[AwardReaction(self._state, reaction) for reaction in comment.reactions],
                    author=self._state.get_partial_user(comment.author_id64),
                    owner=self,
                )
                if after < comment.created_at < before:
                    if limit is not None and yielded >= limit:
                        break
                    comments.append(comment)
                    yielded += 1
                else:
                    break

            count -= len(comments)
            return comments

        comments = await get_comments(min(limit or 100, 100))
        for comment, author in zip(
            comments, await self._state._maybe_users(comment.author.id64 for comment in comments)
        ):
            comment.author = author
            yield comment

        assert count is not None
        while count > 0:
            comments = await get_comments(min(limit or 100, 100))
            for comment, author in zip(
                comments, await self._state._maybe_users(comment.author.id64 for comment in comments)
            ):
                comment.author = author
                yield comment


class _AwardableType(IntEnum):
    # from https://github.com/SteamDatabase/SteamTracking/blob/9ed188c12e9a3474416972873601185c6e0848c7/store.steampowered.com/public/javascript/applications/store/reviewaward.js#L453-L466
    # yes I know it's not great, I thought I had a better one but until I find that, suffer.
    Review = 1
    PublishedFile = 2
    Post = 3
    Topic = 4
    Comment = 5


IDT = TypeVar("IDT", bound=int, default=int, covariant=True)


class Awardable(Protocol[IDT]):  # type: ignore  # ReadOnly (PEP 705) should save this
    """A mixin that implements award functionality."""

    __slots__ = ()

    _state: ConnectionState
    id: ReadOnly[IDT]

    @classproperty
    def _AWARDABLE_TYPE(cls: type[Self]) -> _AwardableType:
        try:
            return _AwardableType[cls.__name__]
        except KeyError:
            raise AttributeError(f"{cls.__name__} is not a valid awardable type")

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        assert cls.__name__ in _AwardableType.__members__, f"{cls.__name__} is not a valid awardable type"

    async def award(self, award: Award, /) -> None:
        """Add an :class:`Award` to this piece of user generated content.

        Parameters
        ----------
        award
            The award to add.
        """
        await self._state.add_award(self, award)

    # async def fetch_reactions(self) -> list[AwardReaction]:
    #     """Fetch the reactions on this piece of user generated content."""
    #     reactions = await self._state.fetch_award_reactions(self)
    #     return [AwardReaction(self._state, reaction) for reaction in reactions]


@dataclass(slots=True)
class UserInventoryInfo(Generic[UserT]):
    user: UserT
    app: UserInventoryInfoApp
    total_count: int
    trade_permissions: str
    load_failed: bool
    store_vetted: bool
    owner_only: bool
    contexts: list[UserInventoryInfoContext]

    async def inventories(self) -> AsyncGenerator[Inventory[Item[UserT], UserT], None]:
        """An :term:`asynchronous iterator` for accessing a user's full inventory in an app."""
        for context in self.contexts:
            yield await self.user.inventory(App(id=self.app.id), context_id=context.id)


class PartialUser(ID[Literal[Type.Individual]], Commentable):
    def __init__(self, state: ConnectionState, id: Intable):
        super().__init__(id, type=Type.Individual)
        self._state = state

    @property
    def _commentable_kwargs(self) -> _CommentableKwargs:
        return {
            "id64": self.id64,
        }

    @classproperty
    def _COMMENTABLE_TYPE(cls: type[Self]) -> _CommentThreadType:  # type: ignore
        return _CommentThreadType.User

    def _send_message(self, content: str) -> Coroutine[Any, Any, UserMessage]:
        return self._state.send_user_message(self.id64, content)

    def _send_media(self, media: Media) -> Coroutine[Any, Any, None]:
        return self._state.http.send_user_media(self.id64, media)

    async def _send_trade(self, trade: TradeOffer[Asset[PartialUser], Asset[ClientUser], Any], **kwargs: Any) -> None:
        try:
            resp = await self._state.http.send_trade_offer(
                self,
                [item.to_dict() for item in trade.sending],
                [item.to_dict() for item in trade.receiving],
                trade.token,
                trade.message or "",
                **kwargs,
            )
        except HTTPException as e:
            if e.code == Result.Revoked and (
                any(item.owner != self for item in trade.receiving)
                or any(item.owner != self._state.user for item in trade.sending)
            ):
                if sys.version_info >= (3, 11):
                    e.add_note("You've probably sent an item isn't either in your inventory or the user's inventory")
                else:
                    raise ValueError(
                        "You've probably sent an item isn't either in your inventory or the user's inventory"
                    ) from e
            raise e
        trade._has_been_sent = True
        needs_confirmation = resp.get("needs_mobile_confirmation", False)
        trade._update_from_send(self._state, resp, self, active=not needs_confirmation)
        if needs_confirmation:
            for tries in range(5):
                try:
                    await trade.confirm()
                    break
                except ConfirmationError:
                    await asyncio.sleep(tries * 2)
            else:
                raise ConfirmationError("Failed to confirm trade offer")
            trade.state = TradeOfferState.Active

        # make sure the trade is updated before this function returns
        self._state._trades[trade.id] = cast(
            "TradeOffer[Item[User], Item[ClientUser], User]", trade
        )  # it gets upcast to this anyway after wait_for_trade
        await self._state.wait_for_trade(trade.id)
        self._state.dispatch("trade", trade)

    @property
    def trade_url(self) -> URL_:
        """The trade url of the user."""
        return URL.COMMUNITY / "tradeoffer/new/" % {"partner": str(self.id)}

    async def inventory_info(self) -> list[UserInventoryInfo[Self]]:
        """Fetch the inventory info of the user.

        .. source:: UserInventoryInfo
        """
        data = await self._state.http.get_user_inventory_info(self.id64)

        return [
            UserInventoryInfo(
                user=self,
                app=UserInventoryInfoApp(
                    self._state,
                    id=info["appid"],
                    name=info["name"],
                    inventory_logo_url=info["inventory_logo"],
                    icon_url=info["icon"],
                ),
                total_count=info["asset_count"],
                trade_permissions=info["trade_permissions"],
                load_failed=bool(info["load_failed"]),
                store_vetted=bool(info["store_vetted"]),
                owner_only=info["owner_only"],
                contexts=[
                    UserInventoryInfoContext(ContextID(ctx["id"]), ctx["name"], ctx["asset_count"])
                    for ctx in info["rgContexts"].values()
                ],
            )
            for info in data
        ]

    async def inventory(
        self, app: App, /, *, context_id: int | None = None, language: Language | None = None
    ) -> Inventory[Item[Self], Self]:
        """Fetch a user's :class:`~steam.Inventory` for trading.

        Parameters
        -----------
        app
            The app to fetch the inventory for.
        context_id
            The context ID for the inventory normally ``2``.
        language
            The language to fetch the inventory in. If ``None`` will default to the current language.

        Raises
        ------
        :exc:`~steam.Forbidden`
            The user's inventory is private.
        """
        if context_id is None:
            context_id = 6 if app.name == "Steam" and context_id is None else 2
        context_id = ContextID(context_id)
        resp = econ.GetInventoryItemsWithDescriptionsResponse().from_dict(
            await self._state.http.get_user_inventory(self.id64, app.id, context_id, language)
        )
        return Inventory(state=self._state, proto=resp, owner=self, app=app, context_id=context_id, language=language)

    async def inventories(self) -> AsyncGenerator[Inventory[Item[Self], Self], None]:
        """Fetches all the inventories a user has."""
        for inventory_info in await self.inventory_info():
            async for inventory in inventory_info.inventories():
                yield inventory

    async def friends(self) -> Sequence[User]:
        """Fetch the list of the users friends."""
        friends = await self._state.http.get_friends_ids(self.id64)
        return await self._state._maybe_users(friends)

    async def apps(self, *, include_free: bool = True) -> list[UserApp]:
        r"""Fetches the :class:`~steam.App`\s the user owns.

        Parameters
        ----------
        include_free
            Whether to include free apps in the list. Defaults to ``True``.
        """
        apps = await self._state.fetch_user_apps(self.id64, include_free)
        return [UserApp(self._state, app) for app in apps]

    async def recently_played(self) -> list[UserRecentlyPlayedApp]:
        """Fetches the apps the user has recently played."""
        apps = await self._state.http.get_user_recently_played_apps(self.id64)
        return [UserRecentlyPlayedApp(self._state, app) for app in apps]

    async def app_stats(self, app: App, /, *, language: Language | None = None) -> UserAppStats:
        """Fetch the stats for the user in the app.

        Parameters
        ----------
        language
            The language to fetch the stats in. If ``None`` will default to the current language.
        """
        msg = await self._state.fetch_user_app_stats(self.id64, app.id)
        schema = VDF_BINARY_LOADS(msg.schema)
        data = cast("achievement.UserAppStats", schema[str(app.id)])
        return UserAppStats(self._state, app, msg, data, language or self._state.language)

    # TODO find a better way to do this
    async def achievements(self, app: App, /, *, language: Language | None = None) -> list[UserAppAchievement]:
        """Fetch the achievements for the user in the app.

        Parameters
        ----------
        language
            The language to fetch the stats in. If ``None`` will default to the current language.
        """
        stats = await self.app_stats(app, language=language)
        return stats.achievements

    async def wishlist(self) -> list[WishlistApp]:
        r"""Get the :class:`.WishlistApp`\s the user has on their wishlist."""
        return [
            WishlistApp(self._state, id=app_id, data=data)
            async for app_id, data in self._state.http.get_user_wishlist(self.id64)
        ]

    async def clans(self, *, auto_chunk: bool = False) -> list[Clan]:
        r"""Fetches a list of :class:`~steam.Clan`\s the user is in.

        Parameters
        ----------
        auto_chunk
            Whether to automatically chunk the clans that are fetched. Defaults to ``False``.
        """

        async def getter(id64: ID64) -> Clan:
            try:
                return await self._state.fetch_clan(id64, maybe_chunk=auto_chunk)
            except WSException as exc:
                if exc.code == Result.RateLimitExceeded:
                    await asyncio.sleep(20)
                    return await getter(id64)
                raise

        return await asyncio.gather(*map(getter, await self._state.http.get_user_clans(self.id64)))

    async def bans(self) -> Ban:
        r"""Fetches the user's :class:`.Ban`\s."""
        (ban,) = await self._state.http.get_user_bans(self.id64)
        return Ban(ban)

    async def is_banned(self) -> bool:
        """Specifies if the user is banned from any part of Steam.

        Shorthand for:

        .. code:: python

            bans = await user.bans()
            bans.is_banned()
        """
        (ban,) = await self._state.http.get_user_bans(self.id64)
        return any((ban["community_banned"], ban["economy_ban"] != "none", ban["vac_banned"]))

    async def level(self) -> int:
        """Fetches the user's level."""
        levels = await self._state.fetch_user_levels(self.id)
        return levels[self.id]

    async def badges(self) -> UserBadges[Self]:
        r"""Fetches the user's :class:`.UserBadges`\s."""
        data = await self._state.http.get_user_badges(self.id64)
        return UserBadges(self._state, self, data=data)

    async def favourite_badge(self) -> FavouriteBadge[PartialApp | Literal[STEAM]] | None:
        """The user's favourite badge."""
        badge = await self._state.fetch_user_favourite_badge(self.id64)
        if not badge.has_favorite_badge:
            return

        return FavouriteBadge(
            self._state,
            id=badge.badgeid,
            app=PartialApp(self._state, id=badge.appid) if badge.appid else STEAM,
            owner=self,
            level=badge.level,
            community_item_id=AssetID(badge.communityitemid),
            type=badge.item_type,
            border_colour=badge.border_color,
        )

    async def equipped_profile_items(self, *, language: Language | None = None) -> EquippedProfileItems[Self]:
        """The user's equipped profile items.

        Parameters
        ----------
        language
            The language to fetch the profile items in. If ``None`` the current language is used
        """
        items = await self._state.fetch_user_equipped_profile_items(self.id64, language)
        return EquippedProfileItems(
            background=ProfileItem(self._state, self, items.profile_background) if items.profile_background else None,
            mini_profile_background=(
                ProfileItem(self._state, self, items.mini_profile_background) if items.mini_profile_background else None
            ),
            avatar_frame=ProfileItem(self._state, self, items.avatar_frame) if items.avatar_frame else None,
            animated_avatar=ProfileItem(self._state, self, items.animated_avatar) if items.animated_avatar else None,
            modifier=ProfileItem(self._state, self, items.profile_modifier) if items.profile_modifier else None,
        )

    async def profile_customisation_info(self) -> ProfileCustomisation[Self]:
        """Fetch a user's profile customisation information."""
        info = await self._state.fetch_user_profile_customisation(self.id64)
        return ProfileCustomisation(self._state, self, info)

    async def profile(self, *, language: Language | None = None) -> Profile[Self]:
        """Fetch a user's entire profile information.

        Parameters
        ----------
        language
            The language to fetch the profile items in. If ``None`` the current language is used

        Note
        ----
        This calls all the profile related functions to return a Profile object which has all the info set.
        """

        return Profile(
            *await asyncio.gather(
                self.equipped_profile_items(language=language),
                self.profile_customisation_info(),
            )
        )

    async def reviews(
        self,
        *,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[Review, None]:
        """An :term:`asynchronous iterator` for accessing a user's :class:`~steam.Review`\\s.

        Examples
        --------
        Usage:

        .. code:: python

            async for review in user.reviews(limit=10):
                recommended = "recommended" if review.recommend else "doesn't recommend"
                print(f"Author: {review.author} {recommended} {review.app}")

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of reviews to search through. Setting this to ``None`` will fetch all the
            user's reviews.
        before
            A time to search for reviews before.
        after
            A time to search for reviews after.

        Yields
        ------
        :class:`~steam.Review`
        """
        from .review import Review
        from .user import ClientUser, User

        pages = 1
        after = after or UNIX_EPOCH
        before = before or DateTime.now()
        yielded = 0
        if type(self) is PartialUser:
            self = await self._state._maybe_user(self.id64)
        assert isinstance(self, (User, ClientUser))

        async def get_reviews(page_number: int = 1) -> AsyncGenerator[Review, None]:
            nonlocal yielded, pages

            # ideally I'd like to find an actual api for these
            page = await self._state.http.get(f"{self.community_url}/recommended", params={"p": page_number})
            soup = BeautifulSoup(page, HTML_PARSER)
            if pages == 1:
                *_, pages = [1] + [int(a["href"].removeprefix("?p=")) for a in soup.find_all("a", class_="pagelink")]
            app_ids = [
                AppID(int(URL_(review.find("div", class_="leftcol").a["href"]).parts[-1]))
                for review in soup.find_all("div", class_="review_box_content")
            ]

            for review_ in await self._state.fetch_user_reviews(self.id64, app_ids):
                review = Review._from_proto(self._state, review_, self)
                if not after < review.created_at < before:
                    return
                if limit is not None and yielded >= limit:
                    return

                yield review
                yielded += 1

        async for review in get_reviews():
            yield review

        for page in range(2, pages + 1):
            async for review in get_reviews(page):
                yield review

    async def fetch_review(self, app: App) -> Review:
        """Fetch this user's review for an app.

        Parameters
        ----------
        app
            The apps to fetch the reviews for.
        """
        (review,) = await self.fetch_reviews(app)
        return review

    async def fetch_reviews(self, *apps: App) -> list[Review]:
        """Fetch this user's review for apps.

        Parameters
        ----------
        apps
            The apps to fetch the reviews for.
        """
        from .review import Review
        from .user import ClientUser, User

        reviews = await self._state.fetch_user_reviews(self.id64, (app.id for app in apps))
        if type(self) is PartialUser:
            # this kinda sucks but it's the best thing I can think of until intersections are added
            self = await self._state.fetch_user(self.id64)
        assert isinstance(self, (User, ClientUser))
        return [Review._from_proto(self._state, review, self) for review in reviews]

    async def fetch_published_file(
        self,
        id: int,
        /,
        *,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
    ) -> PublishedFile[Self] | None:
        """Fetch a published file by its id.

        Parameters
        ----------
        id
            The id of the published file to fetch.
        revision
            The desired revision of the published file to fetch.
        language
            The language to fetch the published file in. If ``None``, the current language is used.
        """
        (file,) = await self._state.fetch_published_files_with_author((PublishedFileID(id),), self, revision, language)
        return file

    async def published_files(
        self,
        *,
        app: App | None = None,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        type: PublishedFileType = PublishedFileType.Community,
        language: Language | None = None,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[PublishedFile[Self], None]:
        """An :term:`asynchronous iterator` for accessing a user's :class:`~steam.PublishedFile`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for file in user.published_files(limit=10):
                print("Author:", file.author, "Published:", file.name)

        All parameters are optional.

        Parameters
        ----------
        app
            The app to fetch published files in.
        type
            The type of published file to fetch.
        revision
            The desired revision of the published file to fetch.
        language
            The language to fetch the published file in. If ``None``, the current language is used.
        limit
            The maximum number of published files to search through. Setting this to ``None`` will fetch all of the
            user's published files.
        before
            A time to search for published files before.
        after
            A time to search for published files after.

        Yields
        ------
        :class:`~steam.PublishedFile`
        """
        from .published_file import PublishedFile

        before = before or DateTime.now()
        after = after or UNIX_EPOCH
        app_id = app.id if app else AppID(0)
        total = 30
        yielded = 0

        while yielded < total:
            page = yielded // 30 + 1
            msg = await self._state.fetch_user_published_files(self.id64, app_id, page, type, revision, language)
            if msg.total:
                total = msg.total

            for file in msg.publishedfiledetails:
                file = PublishedFile(self._state, file, self)
                if not after < file.created_at < before:
                    return
                if limit is not None and yielded >= limit:
                    return
                yield file
                yielded += 1

    async def fetch_post(self, id: int, /) -> Post[Self]:
        """Fetch a post by its id.

        Parameters
        ----------
        id
            The id of the post to fetch.
        """
        from .post import Post

        post = await self._state.fetch_user_post(self.id64, PostID(id))
        return Post(
            self._state,
            PostID(post.postid),
            post.status_text,
            self,
            PartialApp(self._state, id=post.appid) if post.appid else None,
        )

    async def invite_to(self, app: App) -> AppInvite: ...

    def is_friend(self) -> bool:
        """Whether the user is in the :class:`ClientUser`'s friends."""
        return self.id in self._state.user._friends


class BaseUser(PartialUser):
    """An ABC that details the common operations on a Steam user.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClientUser`
        - :class:`~steam.Friend`
        - :class:`~steam.Member`

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.
    """

    __slots__ = ()

    name: str
    """The user's username."""
    last_logoff: datetime | None
    """The last time the user logged off steam. Could be None (e.g. if they are currently online)."""
    last_logon: datetime | None
    """The last time the user logged into steam."""
    last_seen_online: datetime | None
    """The last time the user was seen online."""
    app: PartialApp[str] | None
    """The app the user is currently in. Is ``None`` if the user isn't in an app."""
    state: PersonaState
    """The current persona state of the account (e.g. LookingToTrade)."""
    flags: PersonaStateFlag
    """The persona state flags of the account."""
    rich_presence: dict[str, str] | None
    """The user's rich presence."""
    game_server_ip: IPv4Address | None
    """The IP address of the game server the user is currently playing on."""
    game_server_port: int | None
    """The port of the game server the user is currently playing on."""
    _avatar_sha: bytes
    _state: ConnectionState

    def __repr__(self) -> str:
        attrs = ("name", "state", "id", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    @property
    def mention(self) -> str:
        """The string used to mention the user in chat."""
        return f"[mention={self.id}]@{self.name}[/mention]"

    @property
    def avatar(self) -> Avatar:
        """The user's avatar."""
        return Avatar(self._state, self._avatar_sha)

    async def server(self) -> GameServer:
        """Fetch the game server this user is currently playing on."""
        if self.game_server_ip is None:
            raise ValueError("User is not playing on a game server")
        server = await self._state.client.fetch_server(ip=self.game_server_ip, port=self.game_server_port)
        assert server is not None
        return server


@runtime_checkable
class Messageable(Protocol[MessageT]):
    """An ABC that details the common operations on a Steam message.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClanChannel`
        - :class:`~steam.GroupChannel`
        - :class:`~steam.UserChannel`
    """

    __slots__ = ()
    _state: ConnectionState

    @abc.abstractmethod
    def _send_message(self, content: str) -> Coroutine[Any, Any, MessageT]:
        raise NotImplementedError

    @abc.abstractmethod
    def _send_media(self, media: Media) -> Coroutine[Any, Any, None]:
        raise NotImplementedError

    async def send(self, content: Any = MISSING, /, *, media: Media | None = None) -> MessageT | None:
        """Send a message to a certain destination.

        Parameters
        ----------
        content
            The content of the message to send.
        media
            The media to send to the user.

        Note
        ----
        Anything as passed to ``content`` is implicitly cast to a :class:`str`.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.

        Returns
        -------
        The sent message, only applicable if ``content`` is passed.
        """
        message = None if content is MISSING else await self._send_message(str(content))
        if media is not None:
            await self._send_media(media)

        return message

    @abc.abstractmethod
    async def history(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[MessageT, None]:
        """An :term:`asynchronous iterator` for accessing a channel's :class:`steam.Message`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for message in channel.history(limit=10):
                print("Author:", message.author, "Said:", message.content)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of messages to search through. Setting this to ``None`` will fetch all of the channel's
            messages, but this will be a very slow operation.
        before
            A time to search for messages before.
        after
            A time to search for messages after.

        Yields
        ------
        :class:`~steam.Message`
        """
        raise NotImplementedError
        yield

    async def fetch_message(self, id: int, /) -> MessageT | None:
        """Fetch a message by its :attr:`Message.id`."""
        created_at = STEAM_EPOCH + timedelta(seconds=id >> 32)
        ordinal = id & 0xFFFF
        return await utils.get(
            self.history(
                before=created_at + timedelta(seconds=1),
                after=created_at - timedelta(seconds=1),
            ),
            ordinal=ordinal,
            created_at=created_at,
        )


ClanT = TypeVar("ClanT", bound="Clan | None", default="Clan | None", covariant=True)
GroupT = TypeVar("GroupT", bound="Group | None", default="Group | None", covariant=True)


@dataclass(slots=True)
class Channel(Messageable[MessageT], Generic[MessageT, ClanT, GroupT]):
    _state: ConnectionState
    clan: ClanT = cast(ClanT, None)
    """The clan this channel belongs to."""
    group: GroupT = cast(GroupT, None)
    """The group this channel belongs to."""


def _clean_up_content(content: str) -> str:  # steam does weird stuff with content
    return content.replace(r"\[", "[").replace("\\\\", "\\")


ChannelT = TypeVar("ChannelT", bound=Channel, default=Channel, covariant=True)


class Message(Generic[UserT, ChannelT], metaclass=abc.ABCMeta):
    """Represents a message from a :class:`~steam.User`. This is a base class from which all messages inherit.

    The following classes implement this:

        - :class:`~steam.UserMessage`
        - :class:`~steam.GroupMessage`
        - :class:`~steam.ClanMessage`

    .. container:: operations

        .. describe:: x == y

            Checks if two messages are equal

        .. describe:: hash(x)

            Returns the hash of a message.

    """

    __slots__ = (
        "author",
        "content",
        "channel",
        "clean_content",
        "created_at",
        "ordinal",
        "group",
        "clan",
        "reactions",
        "partial_reactions",
        "_state",
    )

    author: UserT
    """The message's author."""
    created_at: datetime
    """The time this message was sent at."""

    def __init__(self, channel: ChannelT, proto: Any):
        self._state: ConnectionState = channel._state
        self.channel = channel
        """The channel the message was sent in."""
        self.group = channel.group
        """The group the message was sent in. ``None`` if the message wasn't sent in a :class:`~steam.Group`."""
        self.clan = channel.clan
        """The clan the message was sent in. ``None`` if the message wasn't sent in a :class:`~steam.Clan`."""
        self.content = parse_bb_code(proto.message)
        """The message's content.

        Note
        ----
        This is **not** what you will see in the steam client see :attr:`clean_content` for that.
        """
        self.ordinal: int = proto.ordinal
        """A per-channel incremented integer up to ``1000`` for every message sent in a second window."""
        self.clean_content: str = getattr(proto, "message_no_bbcode", "") or _clean_up_content(self.content)
        """The message's clean content without BBCode."""
        self.reactions: list[MessageReaction] = []
        """The message's reactions."""
        self.partial_reactions: list[PartialMessageReaction] = []
        """The message's partial reactions."""

    def __repr__(self) -> str:
        attrs = ("author", "id", "channel")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.channel == other.channel and self.id == other.id

    def __hash__(self) -> int:
        return hash((self.channel, self.id))

    @classmethod
    @abc.abstractmethod
    def _from_history(cls, channel: ChannelT, proto: Any) -> Self:  # type: ignore  # it's a constructor so covariance is fine
        raise NotImplementedError

    @property
    def id(self) -> int:
        """A unique identifier for every message sent in a channel.

        Note
        ----
        This is **not** something Steam provides, this is meant to be a simple way to compare messages.
        """
        # a u64 "snowflake-esk" id measuring of the number of seconds passed since Steam's EPOCH and then the
        # "sequence"/ordinal of the message.
        return int((self.created_at - STEAM_EPOCH).total_seconds()) << 32 | self.ordinal

    @abc.abstractmethod
    async def _react(self, emoticon: Emoticon | Sticker, add: bool) -> None:
        raise NotImplementedError

    async def add_emoticon(self, emoticon: Emoticon, /) -> None:
        """Adds an emoticon to this message.

        Parameters
        ----------
        emoticon
            The emoticon to add to this message.
        """
        await self._react(emoticon, True)

    async def remove_emoticon(self, emoticon: Emoticon, /) -> None:
        """Removes an emoticon from this message.

        Parameters
        ----------
        emoticon
            The emoticon to remove from this message.
        """
        await self._react(emoticon, False)

    async def add_sticker(self, sticker: Sticker, /) -> None:
        """Adds a sticker to this message.

        Parameters
        ----------
        sticker
            The sticker to add to this message.
        """
        await self._react(sticker, True)

    async def remove_sticker(self, sticker: Sticker, /) -> None:
        """Adds a sticker to this message.

        Parameters
        ----------
        sticker
            The sticker to remove from this message.
        """
        await self._react(sticker, False)

    @abc.abstractmethod
    async def ack(self) -> None:
        """Acknowledge this message.

        Note
        ----
        This will acknowledge any messages sent before this message
        """
        raise NotImplementedError()
